"""
Collection of non-standard scikit-learn performance metrics for evaluation of
machine learning model predictions.
"""

import numpy as np
import pandas as pd
import scipy.stats as stats
import datetime

from collections import defaultdict

from statsmodels.regression.mixed_linear_model import MixedLM
from statsmodels.tools.tools import add_constant

from sklearn.metrics import make_scorer, accuracy_score, balanced_accuracy_score
from sklearn.model_selection import cross_val_score
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import VotingRegressor
from sklearn.base import RegressorMixin

from macrosynergy.learning.panel_time_series_split import ExpandingKFoldPanelSplit

from typing import Union


def panel_significance_probability(
    y_true: pd.Series, y_pred: Union[pd.Series, np.ndarray]
) -> float:
    """
    Function to create a linear mixed effects model between the ground truth returns and
    the predicted returns, returning the significance of the model slope.
    Period-specific random effects are included in the model to account for
    return cross-sectional correlations.
    This can be passed into sklearn's make_scorer function to be used as a scorer in a
    grid search or cross validation procedure.

    :param <pd.Series> y_true: Pandas series of ground truth labels. These must be
        multi-indexed by cross-section and date. The dates must be in datetime format.
    :param <Union[pd.Series,np.ndarray]> y_pred: Either a pandas series or numpy array
        of predicted targets. This must have the same length as y_true.

    :return <float> significance_prob: 1 - p-value of the regression slope parameter,
        given by the linear mixed effects model.
    """

    # checks
    if not isinstance(y_true, pd.Series):
        raise TypeError("y_true must be a pandas series")

    if not isinstance(y_true.index, pd.MultiIndex):
        raise ValueError("y_true must be multi-indexed.")

    if not isinstance(y_true.index.get_level_values(1)[0], datetime.date):
        raise TypeError("The inner index of y must be datetime.date.")

    if not (len(y_true) == len(y_pred)):
        raise ValueError("y_true and y_pred must have the same length.")

    if np.all(y_true == 0):
        # Sklearn averages each metric over the CV splits.
        # If all the ground truth labels are zero, the regression is invalid due to a
        # singular matrix. Hence, we return zero in this case.
        significance_prob = 0
        return significance_prob

    # regress ground truth against predictions
    X = add_constant(y_pred)
    groups = y_true.index.get_level_values(1)

    # fit model
    re = MixedLM(y_true, X, groups=groups).fit(reml=False)
    pval = re.pvalues.iloc[1]

    return 1 - pval


def regression_accuracy(
    y_true: pd.Series, y_pred: Union[pd.Series, np.ndarray]
) -> float:
    """
    Function to return the accuracy between the signs of the predictions and targets.

    :param <pd.Series> y_true: Pandas series of ground truth labels. These must be
        multi-indexed by cross-section and date. The dates must be in datetime format.
    :param <Union[pd.Series,np.ndarray]> y_pred: Either a pandas series or numpy array
        of predicted targets. This must have the same length as y_true.

    :return <float>: Accuracy between the signs of the predictions and targets.
    """

    # checks
    if not isinstance(y_true, pd.Series):
        raise TypeError("y_true must be a pandas series")

    if not isinstance(y_true.index, pd.MultiIndex):
        raise ValueError("y_true must be multi-indexed.")

    if not isinstance(y_true.index.get_level_values(1)[0], datetime.date):
        raise TypeError("The inner index of y must be datetime.date.")

    if not (len(y_true) == len(y_pred)):
        raise ValueError("y_true and y_pred must have the same length.")

    return accuracy_score(y_true < 0, y_pred < 0)


def regression_balanced_accuracy(
    y_true: pd.Series, y_pred: Union[pd.Series, np.ndarray]
) -> float:
    """
    Function to return the balanced accuracy between the signs
    of the predictions and targets.

    :param <pd.Series> y_true: Pandas series of ground truth labels. These must be
        multi-indexed by cross-section and date. The dates must be in datetime format.
    :param <Union[pd.Series,np.ndarray]> y_pred: Either a pandas series or numpy array
        of predicted targets. This must have the same length as y_true.

    :return <float>: Balanced accuracy between the signs of the predictions and targets.
    """

    # checks
    if not isinstance(y_true, pd.Series):
        raise TypeError("y_true must be a pandas series")

    if not isinstance(y_true.index, pd.MultiIndex):
        raise ValueError("y_true must be multi-indexed.")

    if not isinstance(y_true.index.get_level_values(1)[0], datetime.date):
        raise TypeError("The inner index of y must be datetime.date.")

    if not (len(y_true) == len(y_pred)):
        raise ValueError("y_true and y_pred must have the same length.")

    return balanced_accuracy_score(y_true < 0, y_pred < 0)


def sharpe_ratio(y_true: pd.Series, y_pred: Union[pd.Series, np.ndarray]) -> float:
    """
    Function to return a Sharpe ratio for a strategy where we go long if the predictions
    are positive and short if the predictions are negative.

    :param <pd.Series> y_true: Pandas series of ground truth labels. These must be
        multi-indexed by cross-section and date. The dates must be in datetime format.
    :param <Union[pd.Series,np.ndarray]> y_pred: Either a pandas series or numpy array
        of predicted targets. This must have the same length as y_true.

    :return <float>: Sharpe ratio for the binary strategy.
    """

    # checks
    if not isinstance(y_true, pd.Series):
        raise TypeError("y_true must be a pandas series")

    if not isinstance(y_true.index, pd.MultiIndex):
        raise ValueError("y_true must be multi-indexed.")

    if not isinstance(y_true.index.get_level_values(1)[0], datetime.date):
        raise TypeError("The inner index of y must be datetime.date.")

    if not (len(y_true) == len(y_pred)):
        raise ValueError("y_true and y_pred must have the same length.")

    portfolio_returns = np.where(y_pred > 0, y_true, -y_true)
    average_return = np.mean(portfolio_returns)
    std_return = np.std(portfolio_returns)

    if std_return == 0:
        # Sklearn averages each metric over the CV splits.
        # If the standard deviation is zero, the portfolio returns are constant.
        # So we return zero to ignore this split in the average.
        sharpe_ratio = 0
    else:
        sharpe_ratio = average_return / std_return

    return sharpe_ratio


def sortino_ratio(y_true: pd.Series, y_pred: Union[pd.Series, np.ndarray]) -> float:
    """
    Function to return a Sortino ratio for a strategy where we go long if the predictions
    are positive and short if the predictions are negative.

    :param <pd.Series> y_true: Pandas series of ground truth labels. These must be
        multi-indexed by cross-section and date. The dates must be in datetime format.
    :param <Union[pd.Series,np.ndarray]> y_pred: Either a pandas series or numpy array
        of predicted targets. This must have the same length as y_true.

    :return <float>: Sortino ratio for the binary strategy.
    """

    # checks
    if not isinstance(y_true, pd.Series):
        raise TypeError("y_true must be a pandas series")

    if not isinstance(y_true.index, pd.MultiIndex):
        raise ValueError("y_true must be multi-indexed.")

    if not isinstance(y_true.index.get_level_values(1)[0], datetime.date):
        raise TypeError("The inner index of y must be datetime.date.")

    if not (len(y_true) == len(y_pred)):
        raise ValueError("y_true and y_pred must have the same length.")

    portfolio_returns = np.where(y_pred > 0, y_true, -y_true)
    negative_returns = portfolio_returns[portfolio_returns < 0]
    average_return = np.mean(portfolio_returns)
    denominator = np.sqrt(np.mean(negative_returns**2))

    if denominator == 0:
        # Sklearn averages each metric over the CV splits.
        # If the denominator is zero, the sortino ratio over this period is invalid.
        # So we return zero to ignore this split in the average.
        sortino_ratio = 0
    else:
        sortino_ratio = average_return / denominator

    return sortino_ratio


def neg_mean_abs_corr(
    estimator: RegressorMixin,
    X_test: pd.DataFrame,
    y_test: Union[pd.DataFrame, pd.Series],
    correlation: str = "pearson",
):
    """
    Custom sklearn scorer for beta estimation to return the negative mean
    absolute correlation between benchmark returns and computed hedged returns over a panel,
    based on estimated beta for each cross-section, with mean taken over the cross-sections.
    X_test is an out-of-sample panel with a single column comprising market returns, whilst
    y_test is an out-of-sample panel with a single column comprising financial contract returns.
    Correlation can be 'pearson', 'spearman' or 'kendall'.

    The estimator is expected to be a seemingly unrelated linear regression model
    possessing a coefs_ dictionary with cross-sectional keys and values corresponding to
    betas.

    Specifically, for each cross-section c in X_test, we compute hedged returns,
    denoted hedged_returns_{c}. Then, we calculate the absolute correlation between each
    hedged_returns_{c} and the benchmark returns.
    Finally, we return:
    $\\text{neg_mean_abs_corr} = - (1/C)\\sum_{c=1}^{C} \\left [ abs_corr_{c} \\right ]$,
    where $C$ is the number of available cross-sections in the panel.

    For calculation of hedged returns for cross-section $c$,
    let $\\hat{\\beta}_{c}$ denote the estimated beta for cross-section c.
    Then $hedged_returns_{c} = contract_returns_{c} - \\hat{\\beta}_{c} * benchmark_returns$.

    .. note::

      This scorer is still **experimental** for now: the predictions
      and the API might change without any deprecation cycle.

    :param <RegressorMixin> estimator: A fitted seemingly unrelated scikit-learn regressor
        with a coefs_ dictionary of type Dict[str, float] containing estimated betas
        for each cross-section.
    :param <pd.DataFrame> X_test: Out-of-sample panel with a single column of market returns.
    :param <Union[pd.DataFrame, pd.Series]> y_test: Out-of-sample panel with a single column
        of financial contract returns.
    :param <str> correlation: The type of correlation to compute. Can be 'pearson',
        'spearman' or 'kendall'. Default is 'pearson'.
    
    :return <float> neg_mean_abs_corr: The negative mean absolute correlation between
        out-of-sample benchmark returns and computed out-of-sample hedged returns
        over the cross-sections in the panel.
        
    NB: this scorer is a specialized function for use in beta estimation.
    """
    market_returns = X_test.iloc[:, 0].copy()
    contract_returns = y_test.copy()
    test_cross_sections = X_test.index.get_level_values(
        0
    ).unique()  # Need the test cids to know which models to evaluate

    # Handle voting regressor case later
    if isinstance(estimator, VotingRegressor):
        estimators = estimator.estimators_
        coefs_list = [est.coefs_ for est in estimators]
        sum_dict = defaultdict(lambda: [0, 0])

        for coefs in coefs_list:
            for key, value in coefs.items():
                sum_dict[key][0] += value
                sum_dict[key][1] += 1

        estimated_coefs = {key: sum / count for key, (sum, count) in sum_dict.items()}
    else:
        estimated_coefs = estimator.coefs_

    running_sum = 0
    xs_count = 0
    for cross_section in test_cross_sections:
        # check whether a model for this cross-section was estimated previously
        if cross_section in estimated_coefs.keys():
            xs_count += 1
            # Get cross-section returns and matched market returns
            contract_returns_c = contract_returns.xs(cross_section)
            market_returns_c = market_returns.xs(
                cross_section
            )  # These were previously replicated across cross sections
            # Calculate hedged returns for this cross-section
            hedged_returns_c = (
                contract_returns_c - estimated_coefs[cross_section] * market_returns_c
            )
            # Compute negative absolute market correlation
            if correlation == "pearson":
                abs_corr = abs(stats.pearsonr(hedged_returns_c, market_returns_c)[0])
            elif correlation == "spearman":
                abs_corr = abs(stats.spearmanr(hedged_returns_c, market_returns_c)[0])
            else:
                # kendall
                abs_corr = abs(stats.kendalltau(hedged_returns_c, market_returns_c)[0])
            # Update running sum
            running_sum += abs_corr
        else:
            # Then a model wasn't estimated for this cross-section
            continue

    if xs_count == 0:
        return np.nan
    else:
        return -running_sum / xs_count


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CPI", "GROWTH", "RIR"]

    df_cids = pd.DataFrame(index=cids, columns=["earliest", "latest"])
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31"]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31"]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31"]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31"]

    tuples = []

    for cid in cids:
        # get list of all elidgible dates
        sdate = df_cids.loc[cid]["earliest"]
        edate = df_cids.loc[cid]["latest"]
        all_days = pd.date_range(sdate, edate)
        work_days = all_days[all_days.weekday < 5]
        for work_day in work_days:
            tuples.append((cid, work_day))

    n_samples = len(tuples)
    ftrs = np.random.normal(loc=0, scale=1, size=(n_samples, 3))
    labels = np.random.normal(loc=0, scale=1, size=n_samples)
    df = pd.DataFrame(
        data=np.concatenate((np.reshape(labels, (-1, 1)), ftrs), axis=1),
        index=pd.MultiIndex.from_tuples(tuples, names=["cid", "real_date"]),
        columns=xcats,
        dtype=np.float32,
    )

    X = df.drop(columns="XR")
    y = df["XR"]

    splitter = ExpandingKFoldPanelSplit(n_splits=4)
    scorer1 = make_scorer(panel_significance_probability, greater_is_better=True)
    scorer2 = make_scorer(sharpe_ratio, greater_is_better=True)
    scorer3 = make_scorer(sortino_ratio, greater_is_better=True)
    scorer4 = make_scorer(regression_accuracy, greater_is_better=True)
    scorer5 = make_scorer(regression_balanced_accuracy, greater_is_better=True)
    cv_results1 = cross_val_score(
        LinearRegression(), X, y, cv=splitter, scoring=scorer1
    )
    cv_results2 = cross_val_score(
        LinearRegression(), X, y, cv=splitter, scoring=scorer2
    )
    cv_results3 = cross_val_score(
        LinearRegression(), X, y, cv=splitter, scoring=scorer3
    )
    cv_results4 = cross_val_score(
        LinearRegression(), X, y, cv=splitter, scoring=scorer4
    )
    cv_results5 = cross_val_score(
        LinearRegression(), X, y, cv=splitter, scoring=scorer5
    )
    print("Probabilities of significances, per split:", cv_results1)
    print("Sharpe ratios, per split:", cv_results2)
    print("Sortino ratios, per split:", cv_results3)
    print("Regression accuracies, per split:", cv_results4)
    print("Regression balanced accuracies, per split:", cv_results5)
    print("Done")



from abc import ABC, abstractmethod
from typing import Union, Optional
import datetime

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.linear_model import LinearRegression, Ridge

from macrosynergy.learning.predictors import LADRegressor

from macrosynergy.management.validation import _validate_Xy_learning


class BaseRegressionSystem(BaseEstimator, RegressorMixin, ABC):

    def __init__(
        self,
        roll: int = None,
        data_freq: str = "D",
        min_xs_samples: int = 2,
    ):
        """
        Base class for systems of regressors.

        :param <int> roll: The lookback of the rolling window for the regression. If None,
            the entire cross-sectional history is used for each regression. This should
            be specified in units of the data frequency, possibly adjusted by the
            data_freq attribute.
        :param <str> data_freq: Training set data frequency. This is primarily
            to be used within the context of market beta estimation in the
            MarketBetaEstimator class in `macrosynergy.learning`. Accepted strings
            are 'D' for daily, 'W' for weekly, 'M' for monthly and 'Q' for quarterly.
            Default is 'D'.
        :param <int> min_xs_samples: The minimum number of samples required in each
            cross-section training set for a regression model to be fitted.
        """
        self.roll = roll
        self.data_freq = data_freq
        self.min_xs_samples = min_xs_samples

        self.models = {}

    def fit(
        self,
        X: pd.DataFrame,
        y: Union[pd.DataFrame, pd.Series],
    ):
        """
        Fit method to fit a regression on each cross-section, subject to
        cross-sectional data availability.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.DataFrame, pd.Series]> y: Pandas series or dataframe of targets
            associated with each sample in X.
        """
        # Fit checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError("The X argument must be a pandas DataFrame.")
        if isinstance(y, np.ndarray):
            # This can happen during sklearn's GridSearch when a voting regressor is used
            y = pd.Series(y, index=X.index)

        _validate_Xy_learning(X, y)

        # Adjust min_xs_samples based on the frequency of the data
        min_xs_samples = self.select_data_freq()

        cross_sections = X.index.get_level_values(0).unique()

        X = self._downsample_by_data_freq(X)
        y = self._downsample_by_data_freq(y)

        for section in cross_sections:
            X_section = X[X.index.get_level_values(0) == section]
            y_section = y[y.index.get_level_values(0) == section]
            unique_dates = sorted(X_section.index.unique())
            num_dates = len(unique_dates)
            # Check if there are enough samples to fit a model
            if not self._check_xs_dates(min_xs_samples, num_dates):
                continue
            # If a roll is specified, then adjust the dates accordingly
            # Only do this if the number of dates is greater than the roll
            if self.roll:
                if num_dates < self.roll:
                    continue
                else:
                    X_section, y_section = self.roll_dates(
                        self.roll, X_section, y_section, unique_dates
                    )
            # Fit the model
            self._fit_cross_section(section, X_section, y_section)

        return self

    def _fit_cross_section(self, section, X_section, y_section):
        """
        Fit a regression model on a single cross-section.
        """
        model = self.create_model()
        model.fit(pd.DataFrame(X_section), y_section)
        # Store model and coefficients
        self.models[section] = model
        self.store_model_info(section, model)

    def predict(
        self,
        X: pd.DataFrame,
    ):
        """
        Predict method to make model predictions over a panel based on the fitted
        seemingly unrelated regression.

        :param <pd.DataFrame> X: Pandas dataframe of input features.

        :return <pd.Series>: Pandas series of predictions, multi-indexed by cross-section
            and date.
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError("The X argument must be a pandas DataFrame.")
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")

        predictions = pd.Series(index=X.index, data=np.nan)

        # Check whether each test cross-section has an associated model
        cross_sections = predictions.index.get_level_values(0).unique()
        for idx, section in enumerate(cross_sections):
            if section in self.models.keys():
                # If a model exists, return the estimated OOS contract return.
                predictions[predictions.index.get_level_values(0) == section] = (
                    self.models[section].predict(X.xs(section, level=0)).flatten()
                )

        return predictions

    def roll_dates(self, roll, X_section, y_section, unique_dates):
        right_dates = unique_dates[-roll:]
        mask = X_section.index.isin(right_dates)
        X_section = X_section[mask]
        y_section = y_section[mask]
        return X_section, y_section

    def select_data_freq(self):
        if self.data_freq == "D":
            min_xs_samples = self.min_xs_samples
        elif self.data_freq == "W":
            min_xs_samples = self.min_xs_samples // 5
        elif self.data_freq == "M":
            min_xs_samples = self.min_xs_samples // 21
        elif self.data_freq == "Q":
            min_xs_samples = self.min_xs_samples // 63
        else:
            raise ValueError(
                "Invalid data frequency. Accepted values are 'D', 'W', 'M' and 'Q'."
            )
        return min_xs_samples

    @abstractmethod
    def store_model_info(self, section, model):
        pass

    @abstractmethod
    def create_model(self):
        """
        Method use to instantiate a regression model for a given cross-section.

        Must be overridden.
        """
        pass

    def _check_xs_dates(self, min_xs_samples, num_dates):
        if num_dates < min_xs_samples:
            return False
        return True

    def _downsample_by_data_freq(self, df):
        return (
            df.groupby(
                [
                    pd.Grouper(level="cid"),
                    pd.Grouper(level="real_date", freq=self.data_freq),
                ]
            )
            .sum()
            .copy()
        )


class LinearRegressionSystem(BaseRegressionSystem):
    """
    Custom scikit-learn predictor class to create a linear OLS seemingly unrelated
    regression model. This means that separate regressions are estimated for each
    cross-section, but evaluation is performed over the panel. Consequently, the results of
    a hyperparameter search will choose a single set of hyperparameters for all cross-sections,
    but the model parameters themselves may differ across cross-sections.

    .. note::

      This estimator is still **experimental** for now: the predictions
      and the API might change without any deprecation cycle.
    """

    def __init__(
        self,
        roll: int = None,
        fit_intercept: bool = True,
        positive: bool = False,
        data_freq: str = "D",
        min_xs_samples: int = 2,
    ):
        """
        Initializes a rolling seemingly unrelated OLS linear regression model. Since
        separate models are estimated for each cross-section, a minimum constraint on the
        number of samples per cross-section, called min_xs_samples, is required for
        sensible inference.

        :param <int> roll: The lookback of the rolling window for the regression. If None,
            the entire cross-sectional history is used for each regression. This should
            be specified in units of the data frequency, possibly adjusted by the
            data_freq attribute.
        :param <bool> fit_intercept: Boolean indicating whether or not to fit intercepts
            for each regression.
        :param <bool> positive: Boolean indicating whether or not to enforce positive
            coefficients for each regression.
        :param <str> data_freq: Training set data frequency. This is primarily
            to be used within the context of market beta estimation in the
            MarketBetaEstimator class in `macrosynergy.learning`. Accepted strings
            are 'D' for daily, 'W' for weekly, 'M' for monthly and 'Q' for quarterly.
            Default is 'D'.
        :param <int> min_xs_samples: The minimum number of samples required in each
            cross-section training set for a regression model to be fitted.
        """
        # Checks
        self._check_init_params(
            roll, fit_intercept, positive, data_freq, min_xs_samples
        )

        self.roll = roll
        self.fit_intercept = fit_intercept
        self.positive = positive
        self.data_freq = data_freq
        self.min_xs_samples = min_xs_samples

        # Create data structures to store model information for each cross-section
        self.coefs_ = {}
        self.intercepts_ = {}

        super().__init__(
            roll=self.roll,
            data_freq=data_freq,
            min_xs_samples=min_xs_samples,
        )

    def create_model(self):
        return LinearRegression(
            fit_intercept=self.fit_intercept,
            positive=self.positive,
        )

    def store_model_info(self, section, model):
        self.coefs_[section] = model.coef_[0]
        self.intercepts_[section] = model.intercept_

    def _check_init_params(
        self, roll, fit_intercept, positive, data_freq, min_xs_samples
    ):
        if (roll is not None) and (not isinstance(roll, int)):
            raise TypeError("roll must be an integer or None.")
        if (roll is not None) and (roll <= 0):
            raise ValueError("roll must be a positive integer.")
        if not isinstance(fit_intercept, bool):
            raise TypeError("fit_intercept must be a boolean.")
        if not isinstance(positive, bool):
            raise TypeError("positive must be a boolean.")
        if not isinstance(data_freq, str):
            raise TypeError("data_freq must be a string.")
        if data_freq not in ["D", "W", "M", "Q"]:
            raise ValueError("data_freq must be one of 'D', 'W', 'M' or 'Q'.")
        if not isinstance(min_xs_samples, int):
            raise TypeError("min_xs_samples must be an integer.")
        if min_xs_samples <= 0:
            raise ValueError("min_xs_samples must be a positive integer.")


class LADRegressionSystem(BaseRegressionSystem):
    """
    Custom scikit-learn predictor class to create a linear LAD seemingly unrelated
    regression model. This means that separate regressions are estimated for each
    cross-section, but evaluation is performed over the panel. Consequently, the results of
    a hyperparameter search will choose a single set of hyperparameters for all cross-sections,
    but the model parameters themselves may differ across cross-sections.

    .. note::

      This estimator is still **experimental** for now: the predictions
      and the API might change without any deprecation cycle.
    """

    def __init__(
        self,
        roll: int = None,
        fit_intercept: bool = True,
        positive: bool = False,
        data_freq: str = "D",
        min_xs_samples: int = 2,
    ):
        """
        Initializes a rolling seemingly unrelated LAD linear regression model. Since
        separate models are estimated for each cross-section, a minimum constraint on the
        number of samples per cross-section, called min_xs_samples, is required for
        sensible inference.

        :param <int> roll: The lookback of the rolling window for the regression. If None,
            the entire cross-sectional history is used for each regression.
        :param <bool> fit_intercept: Boolean indicating whether or not to fit intercepts
            for each regression.
        :param <bool> positive: Boolean indicating whether or not to enforce positive
            coefficients for each regression.
        :param <str> data_freq: Training set data frequency. This is primarily
            to be used within the context of market beta estimation in the
            MarketBetaEstimator class in `macrosynergy.learning`. Accpeted strings
            are 'D' for daily, 'W' for weekly, 'M' for monthly and 'Q' for quarterly.
            Default is 'D'.
        :param <int> min_xs_samples: The minimum number of samples required in each
            cross-section training set for a regression model to be fitted.
        """
        # Checks
        self._check_init_params(
            roll, fit_intercept, positive, data_freq, min_xs_samples
        )

        self.roll = roll
        self.fit_intercept = fit_intercept
        self.positive = positive
        self.data_freq = data_freq
        self.min_xs_samples = min_xs_samples

        # Create data structures to store model information for each cross-section
        self.coefs_ = {}
        self.intercepts_ = {}

        super().__init__(
            roll=roll,
            data_freq=data_freq,
            min_xs_samples=min_xs_samples,
        )

    def create_model(self):
        return LADRegressor(
            fit_intercept=self.fit_intercept,
            positive=self.positive,
        )

    def store_model_info(self, section, model):
        self.coefs_[section] = model.coef_[0]
        self.intercepts_[section] = model.intercept_

    def _check_init_params(
        self, roll, fit_intercept, positive, data_freq, min_xs_samples
    ):
        if not isinstance(roll, int) and roll is not None:
            raise TypeError("roll must be an integer or None.")
        if (roll is not None) and (roll <= 0):
            raise ValueError("roll must be a positive integer.")
        if not isinstance(fit_intercept, bool):
            raise TypeError("fit_intercept must be a boolean.")
        if not isinstance(positive, bool):
            raise TypeError("positive must be a boolean.")
        if not isinstance(data_freq, str):
            raise TypeError("data_freq must be a string.")
        if data_freq not in ["D", "W", "M", "Q"]:
            raise ValueError("data_freq must be one of 'D', 'W', 'M' or 'Q'.")
        if not isinstance(min_xs_samples, int):
            raise TypeError("min_xs_samples must be an integer.")
        if min_xs_samples <= 0:
            raise ValueError("min_xs_samples must be a positive integer.")


class RidgeRegressionSystem(BaseRegressionSystem):
    """
    Custom scikit-learn predictor class to create a seemingly unrelated ridge
    regression model. This means that separate regressions, possibly rolling, are estimated for each
    cross-section, but evaluation is performed over the panel. Consequently, the results of
    a hyperparameter search will choose a single set of hyperparameters for all cross-sections,
    but the model parameters themselves may differ across cross-sections.

    .. note::

      This estimator is still **experimental** for now: the predictions
      and the API might change without any deprecation cycle.
    """

    def __init__(
        self,
        roll: int = None,
        alpha: float = 1.0,
        fit_intercept: bool = True,
        positive: bool = False,
        data_freq: str = "D",
        min_xs_samples: int = 2,
        tol: float = 1e-4,
        solver: str = "lsqr",
    ):
        """
        Initializes a rolling seemingly unrelated ridge regression model. Since
        separate models are estimated for each cross-section, a minimum constraint on the
        number of samples per cross-section, called min_xs_samples, is required for
        sensible inference.

        :param <int> roll: The lookback of the rolling window for the regression. If None,
            the entire cross-sectional history is used for each regression.
        :param <float> alpha: Regularization hyperparameter. Greater values specify stronger
            regularization. This must be a value in $[0, np.inf]$. Default is 1.0.
        :param <bool> fit_intercept: Boolean indicating whether or not to fit intercepts
            for each regression.
        :param <bool> positive: Boolean indicating whether or not to enforce positive
            coefficients for each regression.
        :param <str> data_freq: Training set data frequency. This is primarily
            to be used within the context of market beta estimation in the
            MarketBetaEstimator class in `macrosynergy.learning`. Accpeted strings
            are 'D' for daily, 'W' for weekly, 'M' for monthly and 'Q' for quarterly.
            Default is 'D'.
        :param <int> min_xs_samples: The minimum number of samples required in each
            cross-section training set for a regression model to be fitted.
        :param <float> tol: The tolerance for termination. Default is 1e-4.
        :param <str> solver: Solver to use in the computational routines. Options are
            'auto', 'svd', 'cholesky', 'lsqr', 'sparse_cg', 'sag', 'saga' and 'lbfgs'.
            Default is 'lsqr'.
        """
        # Checks
        self._check_init_params(
            roll, alpha, fit_intercept, positive, data_freq, min_xs_samples, tol, solver
        )

        self.roll = roll
        self.alpha = alpha
        self.fit_intercept = fit_intercept
        self.positive = positive
        self.data_freq = data_freq
        self.min_xs_samples = min_xs_samples
        self.tol = tol
        self.solver = solver

        # Create data structures to store model information for each cross-section
        self.coefs_ = {}
        self.intercepts_ = {}

        super().__init__(
            roll=roll,
            data_freq=data_freq,
            min_xs_samples=min_xs_samples,
        )

    def create_model(self):
        return Ridge(
            fit_intercept=self.fit_intercept,
            positive=self.positive,
            alpha=self.alpha,
            tol=self.tol,
            solver=self.solver,
        )

    def store_model_info(self, section, model):
        self.coefs_[section] = model.coef_[0]
        self.intercepts_[section] = model.intercept_

    def _check_init_params(
        self,
        roll,
        alpha,
        fit_intercept,
        positive,
        data_freq,
        min_xs_samples,
        tol,
        solver,
    ):
        if not isinstance(roll, int) and roll is not None:
            raise TypeError("roll must be an integer or None.")
        if roll is not None and roll <= 0:
            raise ValueError("roll must be a positive integer.")
        if not isinstance(alpha, (int, float)):
            raise TypeError("alpha must be either an integer or a float.")
        if alpha < 0:
            raise ValueError("alpha must be non-negative.")
        if not isinstance(fit_intercept, bool):
            raise TypeError("fit_intercept must be a boolean.")
        if not isinstance(positive, bool):
            raise TypeError("positive must be a boolean.")
        if not isinstance(data_freq, str):
            raise TypeError("data_freq must be a string.")
        if data_freq not in ["D", "W", "M", "Q"]:
            raise ValueError("data_freq must be one of 'D', 'W', 'M' or 'Q'.")
        if not isinstance(min_xs_samples, int):
            raise TypeError("min_xs_samples must be an integer.")
        if min_xs_samples <= 0:
            raise ValueError("min_xs_samples must be a positive integer.")
        if not isinstance(tol, (int, float)):
            raise TypeError("tol must be either an integer or a float.")
        if tol <= 0:
            raise ValueError("tol must be a positive number.")
        if solver not in [
            "auto",
            "svd",
            "cholesky",
            "lsqr",
            "sparse_cg",
            "sag",
            "saga",
            "lbfgs",
        ]:
            raise ValueError(
                "solver must be one of 'auto', 'svd', 'cholesky', 'lsqr', 'sparse_cg', 'sag', 'saga' or 'lbfgs'."
            )


class CorrelationVolatilitySystem(BaseRegressionSystem):
    """
    Custom scikit-learn predictor class written specifically to estimate betas for
    financial contracts with respect to a benchmark return series. Since an estimated beta
    can be decomposed into correlation and volatility components, this class aims to estimate
    these separately, allowing for different lookbacks and weighting schemes for both
    components.

    .. note::

      This estimator is still **experimental** for now: the predictions
      and the API might change without any deprecation cycle.
    """

    def __init__(
        self,
        correlation_lookback: Optional[int] = None,
        correlation_type: str = "pearson",
        volatility_lookback: int = 21,
        volatility_window_type: str = "rolling",
        data_freq: str = "D",
        min_xs_samples: int = 2,
    ):
        """
        Initialize CorrelationVolatilitySystem class.

        :param <Optional[int]> correlation_lookback: The lookback period for the correlation
            calculation. This should be in units of the dataset frequency, possibly
            relating to data_freq. Default is None (use all available history).
        :param <str> correlation_type: The type of correlation to be calculated.
            Accepted values are 'pearson', 'kendall' and 'spearman'. Default is 'pearson'.
        :param <int> volatility_lookback: The lookback period for the volatility
            calculation. This should be in units of the dataset frequency,
            possibly relating to data_freq. Default is 21.
        :param <str> volatility_window_type: The type of window to use for the volatility
            calculation. Accepted values are 'rolling' and 'exponential'. Default is 'rolling'.
        :param <str> data_freq: Training set data frequency for downsampling. Default is 'D'.
        :param <int> min_xs_samples: The minimum number of samples required in each
            cross-section training set for a regression model to be fitted.
        """
        # Checks
        self._check_init_params(
            correlation_lookback,
            correlation_type,
            volatility_lookback,
            volatility_window_type,
            data_freq,
            min_xs_samples,
        )

        self.correlation_lookback = correlation_lookback
        self.correlation_type = correlation_type
        self.volatility_lookback = volatility_lookback
        self.volatility_window_type = volatility_window_type
        self.data_freq = data_freq
        self.min_xs_samples = min_xs_samples

        super().__init__(
            roll=None,
            data_freq=data_freq,
            min_xs_samples=min_xs_samples,
        )

        # Create data structures to store the estimated betas for each cross-section
        self.coefs_ = {}

    def _fit_cross_section(self, section, X_section, y_section):
        # Estimate local standard deviations of the benchmark and contract return
        if self.volatility_window_type == "rolling":
            X_section_std = X_section[-self.volatility_lookback :].std().iloc[-1]
            y_section_std = y_section[-self.volatility_lookback :].std()
        elif self.volatility_window_type == "exponential":
            X_section_std = (
                X_section.ewm(span=self.volatility_lookback).std().iloc[-1, 0]
            )
            y_section_std = y_section.ewm(span=self.volatility_lookback).std().iloc[-1]

        # Estimate local correlation between the benchmark and contract return
        if self.correlation_lookback is not None:
            X_section_corr = X_section.tail(self.correlation_lookback)
            y_section_corr = y_section.tail(self.correlation_lookback)
            corr = X_section_corr.corrwith(
                y_section_corr, method=self.correlation_type
            ).iloc[-1]
        else:
            corr = X_section.corrwith(y_section, method=self.correlation_type).iloc[-1]

        # Get beta estimate and store it
        beta = corr * (y_section_std / X_section_std)
        self.store_model_info(section, beta)

    def predict(
        self,
        X: pd.DataFrame,
    ):
        """
        Predict method to make a naive zero prediction for each cross-section. This is
        because the only use of this class is to estimate betas, which were computed
        during the fit method, whose quality can be assessed by a custom scikit-learn metric
        that directly access the betas without the need for a predict method.

        :param <pd.DataFrame> X: Pandas dataframe of input features.

        :return <pd.Series>: Pandas series of zero predictions, multi-indexed by cross-section
            and date.
        """
        predictions = pd.Series(index=X.index, data=0)

        return predictions

    def store_model_info(self, section, beta):
        self.coefs_[section] = beta

    def create_model(self):
        raise NotImplementedError("This method is not implemented for this class.")

    def _check_xs_dates(self, min_xs_samples, num_dates):
        if num_dates < min_xs_samples:
            return False
        # If the correlation lookback is greater than the number of available dates, skip
        # to the next cross-section
        if (
            self.correlation_lookback is not None
            and num_dates < self.correlation_lookback
        ):
            return False
        # If the volatility lookback is greater than the number of available dates, skip
        # to the next cross-section
        if num_dates < self.volatility_lookback:
            return False
        return True

    def _check_init_params(
        self,
        correlation_lookback,
        correlation_type,
        volatility_lookback,
        volatility_window_type,
        data_freq,
        min_xs_samples,
    ):
        if correlation_lookback is not None:
            if not isinstance(correlation_lookback, int):
                raise TypeError("correlation_lookback must be an integer.")
            if correlation_lookback <= 0:
                raise ValueError("correlation_lookback must be a positive integer.")
        if not isinstance(correlation_type, str):
            raise TypeError("correlation_type must be a string.")
        if correlation_type not in ["pearson", "kendall", "spearman"]:
            raise ValueError(
                "correlation_type must be one of 'pearson', 'kendall' or 'spearman'."
            )
        if not isinstance(volatility_lookback, int):
            raise TypeError("volatility_lookback must be an integer.")
        if volatility_lookback <= 0:
            raise ValueError("volatility_lookback must be a positive integer.")
        if not isinstance(volatility_window_type, str):
            raise TypeError("volatility_window_type must be a string.")
        if volatility_window_type not in ["rolling", "exponential"]:
            raise ValueError(
                "volatility_window_type must be one of 'rolling' or 'exponential'."
            )
        if not isinstance(data_freq, str):
            raise TypeError("data_freq must be a string.")
        if data_freq not in ["D", "W", "M", "Q"]:
            raise ValueError("data_freq must be one of 'D', 'W', 'M' or 'Q.")
        if not isinstance(min_xs_samples, int):
            raise TypeError("min_xs_samples must be an integer.")
        if min_xs_samples <= 0:
            raise ValueError("min_xs_samples must be a positive integer.")


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    from sklearn.linear_model import LinearRegression

    import macrosynergy.management as msm
    from macrosynergy.management import make_qdf

    np.random.seed(1)

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "BENCH_XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example: Unbalanced panel """

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2013-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2010-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2010-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2010-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["BENCH_XR"] = ["2010-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2010-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2010-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2010-01-01", "2020-12-31", 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {"GBP": ["2009-01-01", "2012-06-30"], "CAD": ["2018-01-01", "2100-01-01"]}
    dfd = msm.reduce_df(df=dfd, cids=cids, xcats=xcats, blacklist=black)

    dfd = dfd.pivot(index=["cid", "real_date"], columns="xcat", values="value")

    # Demonstration of CorrelationVolatilitySystem usage

    X2 = pd.DataFrame(dfd["BENCH_XR"])
    y2 = dfd["XR"]
    cv = CorrelationVolatilitySystem().fit(X2, y2)
    print(cv.coefs_)

    # Demonstration of LinearRegressionSystem usage
    X1 = dfd.drop(columns=["XR", "BENCH_XR"])
    y1 = dfd["XR"]
    lr = LinearRegressionSystem(data_freq="W").fit(X1, y1)
    print(lr.coefs_)



"""
Utility functions for working with DataFrames.
"""

from macrosynergy.management.types import QuantamentalDataFrame

import warnings
from typing import Iterable, List, Optional, Union

import numpy as np
import pandas as pd
import datetime
import macrosynergy.management.constants as ms_constants
from macrosynergy.management.utils.core import (
    get_cid,
    get_xcat,
    _map_to_business_day_frequency,
    is_valid_iso_date,
)
import functools

IDX_COLS_SORT_ORDER = ["cid", "xcat", "real_date"]


def standardise_dataframe(
    df: pd.DataFrame, verbose: bool = False
) -> QuantamentalDataFrame:
    """
    Applies the standard JPMaQS Quantamental DataFrame format to a DataFrame.

    :param <pd.DataFrame> df: The DataFrame to be standardized.
    :param <bool> verbose: Whether to print warnings if the DataFrame is not in the
        correct format.
    :return <pd.DataFrame>: The standardized DataFrame.
    :raises <TypeError>: If the input is not a pandas DataFrame.
    :raises <ValueError>: If the input DataFrame is not in the correct format.
    """
    idx_cols: List[str] = QuantamentalDataFrame.IndexCols
    metric_columns: List[str] = ms_constants.JPMAQS_METRICS

    # Check if the input DF contains the standard columns
    if not set(df.columns).issuperset(set(idx_cols)):
        fail_str: str = (
            f"Error : Tried to standardize DataFrame but failed."
            f"DataFrame not in the correct format. Please ensure "
            f"that the DataFrame has the following columns: "
            f"'cid', 'xcat', 'real_date', along with any other "
            "variables you wish to include (e.g. 'value', 'mop_lag', "
            "'eop_lag', 'grading')."
        )

        try:
            dft: pd.DataFrame = df.reset_index()
            found_cols: list = dft.columns.tolist()
            fail_str += f"\nFound columns: {found_cols}"
            if not set(dft.columns).issuperset(set(idx_cols)):
                raise ValueError(fail_str)
            df = dft.copy()
        except:
            raise ValueError(fail_str)

        # check if there is atleast one more column
        if len(df.columns) < 4:
            raise ValueError(fail_str)

    # Convert date and ensure specific columns are strings in one step
    # 'datetime64[ns]' is the default dtype for datetime columns in pandas
    df["real_date"] = pd.to_datetime(
        df["real_date"],
        format="%Y-%m-%d",
    ).astype("datetime64[ns]")
    df["cid"] = df["cid"].astype(str)
    df["xcat"] = df["xcat"].astype(str)
    # sort by cid, xcat and real_date to allow viewing stacked timeseries easily
    df = (
        df.drop_duplicates(subset=idx_cols, keep="last")
        .sort_values(by=IDX_COLS_SORT_ORDER)
        .reset_index(drop=True)
    )

    # Sort the 'remaining' columns
    ## No more row-reordering or shape changes after this point

    jpmaqs_metrics = [mtr for mtr in metric_columns if mtr in df.columns]
    non_jpmaqs_metrics = (set(df.columns) - set(jpmaqs_metrics)) - set(idx_cols)
    col_order = idx_cols + jpmaqs_metrics + sorted(list(non_jpmaqs_metrics))
    df = df[col_order]

    # for every remaining col, try to convert to float
    for col in col_order:
        try:
            df[col] = df[col].astype(float)
        except:
            pass

    assert isinstance(df, QuantamentalDataFrame), "Failed to standardize DataFrame"
    return df


def drop_nan_series(
    df: pd.DataFrame, column: str = "value", raise_warning: bool = False
) -> QuantamentalDataFrame:
    """
    Drops any series that are entirely NaNs.
    Raises a user warning if any series are dropped.

    :param <pd.DataFrame> df: The dataframe to be cleaned.
    :param <str> column: The column to be used as the value column, defaults to
        "value".
    :param <bool> raise_warning: Whether to raise a warning if any series are dropped.
    :return <pd.DataFrame | QuantamentalDataFrame>: The cleaned DataFrame.
    :raises <TypeError>: If the input is not a pandas DataFrame.
    :raises <ValueError>: If the input DataFrame is not in the correct format.
    """
    if not isinstance(df, QuantamentalDataFrame):
        raise TypeError("Argument `df` must be a Quantamental DataFrame.")

    if not column in df.columns:
        raise ValueError(f"Column {column} not present in DataFrame.")

    if not df["value"].isna().any():
        return df

    if not isinstance(raise_warning, bool):
        raise TypeError("Error: The raise_warning argument must be a boolean.")

    df_orig: pd.DataFrame = df.copy()
    for cd, xc in df_orig.groupby(["cid", "xcat"]).groups:
        sel_series: pd.Series = df_orig[
            (df_orig["cid"] == cd) & (df_orig["xcat"] == xc)
        ]["value"]
        if sel_series.isna().all():
            if raise_warning:
                warnings.warn(
                    message=f"The series {cd}_{xc} is populated "
                    "with NaNs only, and will be dropped.",
                    category=UserWarning,
                )
            df = df[~((df["cid"] == cd) & (df["xcat"] == xc))]

    return df.reset_index(drop=True)


def qdf_to_ticker_df(df: pd.DataFrame, value_column: str = "value") -> pd.DataFrame:
    """
    Converts a standardized JPMaQS DataFrame to a wide format DataFrame
    with each column representing a ticker.

    :param <pd.DataFrame> df: A standardised quantamental dataframe.
    :param <str> value_column: The column to be used as the value column, defaults to
        "value". If the specified column is not present in the DataFrame, a column named
        "value" will be used. If there is no column named "value", the first
        column in the DataFrame will be used instead.
    :return <pd.DataFrame>: The converted DataFrame.
    """
    if not isinstance(df, QuantamentalDataFrame):
        raise TypeError("Argument `df` must be a QuantamentalDataFrame.")

    if not isinstance(value_column, str):
        raise TypeError("Argument `value_column` must be a string.")

    if not value_column in df.columns:
        cols: List[str] = list(set(df.columns) - set(QuantamentalDataFrame.IndexCols))
        if "value" in cols:
            value_column: str = "value"

        warnings.warn(
            f"Value column specified in `value_column` ({value_column}) "
            f"is not present in the DataFrame. Defaulting to {cols[0]}."
        )
        value_column: str = cols[0]

    return (
        df.assign(ticker=df["cid"] + "_" + df["xcat"])
        .pivot(index="real_date", columns="ticker", values=value_column)
        .rename_axis(None, axis=1)
    )


def ticker_df_to_qdf(df: pd.DataFrame, metric: str = "value") -> QuantamentalDataFrame:
    """
    Converts a wide format DataFrame (with each column representing a ticker)
    to a standardized JPMaQS DataFrame.

    :param <pd.DataFrame> df: A wide format DataFrame.
    :return <pd.DataFrame>: The converted DataFrame.
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError("Argument `df` must be a pandas DataFrame.")
    if not isinstance(metric, str):
        raise TypeError("Argument `metric` must be a string.")

    # pivot to long format
    df = (
        df.stack(level=0).reset_index().rename(columns={0: metric, "level_1": "ticker"})
    )

    df["cid"] = get_cid(df["ticker"])
    df["xcat"] = get_xcat(df["ticker"])
    df = df.drop(columns=["ticker"])

    return standardise_dataframe(df=df)


def concat_single_metric_qdfs(
    df_list: List[QuantamentalDataFrame],
    errors: str = "ignore",
) -> QuantamentalDataFrame:
    """
    Combines a list of Quantamental DataFrames into a single DataFrame.

    :param <List[QuantamentalDataFrame]> df_list: A list of Quantamental DataFrames.
    :param <str> errors: The error handling method to use. If 'raise', then invalid
        items in the list will raise an error. If 'ignore', then invalid items will be
        ignored. Default is 'ignore'.
    :return <QuantamentalDataFrame>: The combined DataFrame.
    """
    if not isinstance(df_list, list):
        raise TypeError("Argument `df_list` must be a list.")

    if errors not in ["raise", "ignore"]:
        raise ValueError("`errors` must be one of 'raise' or 'ignore'.")

    if errors == "raise":
        if not all([isinstance(df, QuantamentalDataFrame) for df in df_list]):
            raise TypeError(
                "All elements in `df_list` must be Quantamental DataFrames."
            )
    else:
        df_list = [df for df in df_list if isinstance(df, QuantamentalDataFrame)]
        if len(df_list) == 0:
            return None

    def _get_metric(df: QuantamentalDataFrame) -> str:
        lx = list(set(df.columns) - set(QuantamentalDataFrame.IndexCols))
        if len(lx) != 1:
            raise ValueError(
                "Each QuantamentalDataFrame must have exactly one metric column."
            )
        return lx[0]

    def _group_by_metric(
        dfl: List[QuantamentalDataFrame], fm: List[str]
    ) -> List[List[QuantamentalDataFrame]]:
        r = [[] for _ in range(len(fm))]
        while dfl:
            metric = _get_metric(df=dfl[0])
            r[fm.index(metric)] += [dfl.pop(0)]
        return r

    found_metrics = list(set(map(_get_metric, df_list)))

    df_list = _group_by_metric(dfl=df_list, fm=found_metrics)

    # use pd.merge to join on QuantamentalDataFrame.IndexCols
    df: pd.DataFrame = functools.reduce(
        lambda left, right: pd.merge(
            left, right, on=["real_date", "cid", "xcat"], how="outer"
        ),
        map(
            lambda fm: pd.concat(df_list.pop(0), axis=0, ignore_index=False),
            found_metrics,
        ),
    )

    return standardise_dataframe(df)


def apply_slip(
    df: QuantamentalDataFrame,
    slip: int,
    cids: Optional[List[str]] = None,
    xcats: Optional[List[str]] = None,
    tickers: Optional[List[str]] = None,
    metrics: List[str] = ["value"],
    raise_error: bool = True,
) -> QuantamentalDataFrame:
    """
    Applies a slip, i.e. a negative lag, to the DataFrame
    for the given cross-sections and categories, on the given metrics.

    :param <QuantamentalDataFrame> target_df: DataFrame to which the slip is applied.
    :param <int> slip: Slip to be applied.
    :param <List[str]> cids: List of cross-sections.
    :param <List[str]> xcats: List of target categories.
    :param <List[str]> metrics: List of metrics to which the slip is applied.
    :return <QuantamentalDataFrame> target_df: DataFrame with the slip applied.
    :raises <TypeError>: If the provided parameters are not of the expected type.
    :raises <ValueError>: If the provided parameters are semantically incorrect.
    """
    if not isinstance(df, QuantamentalDataFrame):
        raise TypeError("Argument `df` must be a QuantamentalDataFrame.")

    df = df.copy()
    if not (isinstance(slip, int) and slip >= 0):
        raise ValueError("Slip must be a non-negative integer.")

    if not any([cids, xcats, tickers]):
        raise ValueError("One of `tickers`, `cids` or `xcats` must be provided.")
    if tickers is not None:
        if cids is not None or xcats is not None:
            raise ValueError("Cannot specify both `tickers` and `cids`/`xcats`.")
    if cids is None:
        cids = df["cid"].unique()
    if xcats is None:
        xcats = df["xcat"].unique()

    if tickers is not None:
        sel_tickers = tickers
    else:
        sel_tickers: List[str] = [f"{cid}_{xcat}" for cid in cids for xcat in xcats]

    df["ticker"] = df["cid"] + "_" + df["xcat"]
    err_str = (
        "Tickers targetted for applying slip are not present in the DataFrame.\n"
        "Missing tickers: {tickers}"
    )

    if not set(sel_tickers).issubset(set(df["ticker"].unique())):
        missing_tickers = sorted(list(set(sel_tickers) - set(df["ticker"].unique())))
        _err_str = err_str.format(tickers=missing_tickers)
        if raise_error:
            raise ValueError(_err_str)
        else:
            warnings.warn(_err_str)

    slip: int = slip.__neg__()

    filtered_df = df[df["ticker"].isin(sel_tickers)]
    filtered_df.loc[:, metrics] = filtered_df.loc[:, metrics].astype(float)
    filtered_df.loc[:, metrics] = filtered_df.groupby("ticker")[metrics].shift(slip)

    df.loc[df["ticker"].isin(sel_tickers), metrics] = filtered_df[metrics]

    df = df.drop(columns=["ticker"]).reset_index(drop=True)
    assert isinstance(df, QuantamentalDataFrame), "Failed to apply slip."

    return df


def downsample_df_on_real_date(
    df: pd.DataFrame,
    groupby_columns: List[str] = [],
    freq: str = "M",
    agg: str = "mean",
):
    """
    Downsample JPMaQS DataFrame.

    :param <pd.Dataframe> df: standardized JPMaQS DataFrame with the necessary columns:
        'cid', 'xcat', 'real_date' and at least one column with values of interest.
    :param <List> groupby_columns: a list of columns used to group the DataFrame.
    :param <str> freq: frequency option. Per default the correlations are calculated
        based on the native frequency of the datetimes in 'real_date', which is business
        daily. Downsampling options include weekly ('W'), monthly ('M'), or quarterly
        ('Q') mean.
    :param <str> agg: aggregation method. Must be one of "mean" (default), "median",
        "min", "max", "first" or "last".

    :return <pd.DataFrame>: the downsampled DataFrame.
    """

    if not set(groupby_columns).issubset(df.columns):
        raise ValueError(
            "The columns specified in 'groupby_columns' were not found in the DataFrame."
        )

    if not isinstance(freq, str):
        raise TypeError("`freq` must be a string")
    else:
        freq: str = _map_to_business_day_frequency(freq)

    if not isinstance(agg, str):
        raise TypeError("`agg` must be a string")
    else:
        agg: str = agg.lower()
        if agg not in ["mean", "median", "min", "max", "first", "last"]:
            raise ValueError(
                "`agg` must be one of 'mean', 'median', 'min', 'max', 'first', 'last'"
            )

    return (
        df.set_index("real_date")
        .groupby(groupby_columns)
        .resample(freq)
        .agg(agg, numeric_only=True)
        .reset_index()
    )


def update_df(df: pd.DataFrame, df_add: pd.DataFrame, xcat_replace: bool = False):
    """
    Append a standard DataFrame to a standard base DataFrame with ticker replacement on
    the intersection.

    :param <pd.DataFrame> df: standardised base JPMaQS DataFrame with the following
        necessary columns: 'cid', 'xcat', 'real_date' and 'value'.
    :param <pd.DataFrame> df_add: another standardised JPMaQS DataFrame, with the latest
        values, to be added with the necessary columns: 'cid', 'xcat', 'real_date', and
        'value'. Columns that are present in the base DataFrame but not in the appended
        DataFrame will be populated with NaN values.
    :param <bool> xcat_replace: all series belonging to the categories in the added
        DataFrame will be replaced, rather than just the added tickers.
        N.B.: tickers are combinations of cross-sections and categories.

    :return <pd.DataFrame>: standardised DataFrame with the latest values of the modified
        or newly defined tickers added.
    """

    # index_cols = ["cid", "xcat", "real_date"]
    # Consider the other possible metrics that the DataFrame could be defined over

    df_cols = set(df.columns)
    df_add_cols = set(df_add.columns)

    error_message = f"The base DataFrame must be a Quantamental Dataframe."
    if not isinstance(df, QuantamentalDataFrame):
        raise TypeError(error_message)

    error_message = f"The added DataFrame must be a Quantamental Dataframe."
    if not isinstance(df_add, QuantamentalDataFrame):
        raise TypeError(error_message)

    error_message = (
        "The two Quantamental DataFrames must share at least "
        "four columns including than 'real_date', 'cid', and 'xcat'."
    )

    all_cols = df_cols.union(df_add_cols)
    if all_cols != df_cols and all_cols != df_add_cols:
        raise ValueError(error_message)

    if not xcat_replace:
        df = update_tickers(df, df_add)

    else:
        df = update_categories(df, df_add)

    # sort same as in `standardise_dataframe`
    return df.sort_values(by=IDX_COLS_SORT_ORDER).reset_index(drop=True)


def update_tickers(df: pd.DataFrame, df_add: pd.DataFrame):
    """
    Method used to update aggregate DataFrame on a ticker level.

    :param <pd.DataFrame> df: aggregate DataFrame used to store all tickers.
    :param <pd.DataFrame> df_add: DataFrame with the latest values.

    """
    df = pd.concat([df, df_add], axis=0, ignore_index=True)

    df = df.drop_duplicates(
        subset=["real_date", "xcat", "cid"], keep="last"
    ).reset_index(drop=True)
    return df


def update_categories(df: pd.DataFrame, df_add):
    """
    Method used to update the DataFrame on the category level.

    :param <pd.DataFrame> df: base DataFrame.
    :param <pd.DataFrame> df_add: appended DataFrame.

    """

    incumbent_categories = list(df["xcat"].unique())
    new_categories = list(df_add["xcat"].unique())

    # Union of both category columns from the two DataFrames.
    append_condition = set(incumbent_categories) | set(new_categories)
    intersect = list(set(incumbent_categories).intersection(set(new_categories)))

    if len(append_condition) == len(incumbent_categories + new_categories):
        df = pd.concat([df, df_add], axis=0, ignore_index=True)

    # Shared categories plus any additional categories previously not defined in the base
    # DataFrame.
    else:
        df = df[~df["xcat"].isin(intersect)]
        df = pd.concat([df, df_add], axis=0, ignore_index=True)

    return df


def reduce_df(
    df: pd.DataFrame,
    xcats: Union[str, List[str]] = None,
    cids: List[str] = None,
    start: str = None,
    end: str = None,
    blacklist: dict = None,
    out_all: bool = False,
    intersect: bool = False,
):
    """
    Filter DataFrame by xcats and cids and notify about missing xcats and cids.

    :param <pd.Dataframe> df: standardized JPMaQS DataFrame with the necessary columns:
        'cid', 'xcat', 'real_date' and 'value'.
    :param <Union[str, List[str]]> xcats: extended categories to be filtered on. Default is
        all in the DataFrame.
    :param <List[str]> cids: cross sections to be checked on. Default is all in the
        dataframe.
    :param <str> start: string representing the earliest date. Default is None.
    :param <str> end: string representing the latest date. Default is None.
    :param <dict> blacklist: cross-sections with date ranges that should be excluded from
        the data frame. If one cross-section has several blacklist periods append numbers
        to the cross-section code.
    :param <bool> out_all: if True the function returns reduced dataframe and selected/
        available xcats and cids.
        Default is False, i.e. only the DataFrame is returned
    :param <bool> intersect: if True only retains cids that are available for all xcats.
        Default is False.

    :return <pd.Dataframe>: reduced DataFrame that also removes duplicates or
        (for out_all True) DataFrame and available and selected xcats and cids.
    """

    if xcats is not None:
        if not isinstance(xcats, list):
            xcats = [xcats]

    if start:
        df = df[df["real_date"] >= pd.to_datetime(start)]

    if end:
        df = df[df["real_date"] <= pd.to_datetime(end)]

    if blacklist is not None:
        for key, value in blacklist.items():
            df = df[
                ~(
                    (df["cid"] == key[:3])
                    & (df["real_date"] >= pd.to_datetime(value[0]))
                    & (df["real_date"] <= pd.to_datetime(value[1]))
                )
            ]

    if xcats is None:
        xcats = sorted(df["xcat"].unique())
    else:
        xcats_in_df = df["xcat"].unique()
        xcats = [xcat for xcat in xcats if xcat in xcats_in_df]

    df = df[df["xcat"].isin(xcats)]

    if intersect:
        cids_in_df = set.intersection(
            *(set(df[df["xcat"] == xcat]["cid"].unique()) for xcat in xcats)
        )
    else:
        cids_in_df = df["cid"].unique()

    if cids is None:
        cids = sorted(cids_in_df)
    else:
        cids = [cids] if isinstance(cids, str) else cids
        cids = [cid for cid in cids if cid in cids_in_df]

    df = df[df["cid"].isin(cids)]

    if out_all:
        return df.drop_duplicates(), xcats, sorted(cids)
    else:
        return df.drop_duplicates()


def reduce_df_by_ticker(
    df: pd.DataFrame,
    ticks: List[str] = None,
    start: str = None,
    end: str = None,
    blacklist: dict = None,
):
    """
    Filter dataframe by xcats and cids and notify about missing xcats and cids

    :param <pd.Dataframe> df: standardized dataframe with the following columns:
        'cid', 'xcat', 'real_date'.
    :param <List[str]> ticks: tickers (cross sections + base categories)
    :param <str> start: string in ISO 8601 representing earliest date. Default is None.
    :param <str> end: string ISO 8601 representing the latest date. Default is None.
    :param <dict> blacklist: cross sections with date ranges that should be excluded from
        the dataframe. If one cross section has several blacklist periods append numbers
        to the cross section code.

    :return <pd.Dataframe>: reduced dataframe that also removes duplicates
    """

    dfx = df.copy()

    if start is not None:
        dfx = dfx[dfx["real_date"] >= pd.to_datetime(start)]

    if end is not None:
        dfx = dfx[dfx["real_date"] <= pd.to_datetime(end)]

    # Blacklisting by cross-section.
    if blacklist is not None:
        masks = []
        for key, value in blacklist.items():
            filt1 = dfx["cid"] == key[:3]
            filt2 = dfx["real_date"] >= pd.to_datetime(value[0])
            filt3 = dfx["real_date"] <= pd.to_datetime(value[1])
            combined_mask = filt1 & filt2 & filt3
            masks.append(combined_mask)

        if masks:
            combined_mask = pd.concat(masks, axis=1).any(axis=1)
            dfx = dfx[~combined_mask]

    dfx["ticker"] = dfx["cid"] + "_" + dfx["xcat"]
    ticks_in_df = dfx["ticker"].unique()
    if ticks is None:
        ticks = sorted(ticks_in_df)
    else:
        ticks = [tick for tick in ticks if tick in ticks_in_df]

    dfx = dfx[dfx["ticker"].isin(ticks)]

    return dfx.drop_duplicates()


def categories_df_aggregation_helper(dfx: pd.DataFrame, xcat_agg: str):
    """
    Helper method to down-sample each category in the DataFrame by aggregating over the
    intermediary dates according to a prescribed method.

    :param <List[str]> dfx: standardised DataFrame defined exclusively on a single
        category.
    :param <List[str]> xcat_agg: associated aggregation method for the respective
        category.

    """

    dfx = dfx.groupby(["xcat", "cid", "custom_date"])
    dfx = dfx.aggregate(xcat_agg, numeric_only=True).reset_index()

    if "real_date" in dfx.columns:
        dfx = dfx.drop(["real_date"], axis=1)
    dfx = dfx.rename(columns={"custom_date": "real_date"})

    return dfx


def _categories_df_explanatory_df(
    dfw: pd.DataFrame,
    explanatory_xcats: List[str],
    agg_method: str,
    sum_condition: bool,
    lag: int,
):
    """
    Produces the explanatory column(s) for the custom DataFrame.

    :param <pd.DataFrame> dfw: group-by DataFrame which has been down-sampled. The
        respective aggregation method will be applied.
    :param <List[str]> explanatory_xcats: list of explanatory category(s).
    :param <str> agg_meth: aggregation method used for all explanatory variables.
    :param <dict> sum_condition: required boolean to negate erroneous zeros if the
        aggregate method used, for the explanatory variable, is sum.
    :param <int> lag: lag of explanatory category(s). Applied uniformly to each
        category.
    """

    dfw_explanatory = pd.DataFrame()
    for xcat in explanatory_xcats:
        if not sum_condition:
            explanatory_col = dfw[xcat].agg(agg_method).astype(dtype=np.float32)
        else:
            explanatory_col = dfw[xcat].sum(min_count=1)

        if lag > 0:
            explanatory_col = explanatory_col.groupby(level=0).shift(lag)

        dfw_explanatory[xcat] = explanatory_col

    return dfw_explanatory


def categories_df(
    df: pd.DataFrame,
    xcats: List[str],
    cids: List[str] = None,
    val: str = "value",
    start: str = None,
    end: str = None,
    blacklist: dict = None,
    years: int = None,
    freq: str = "M",
    lag: int = 0,
    fwin: int = 1,
    xcat_aggs: List[str] = ["mean", "mean"],
):
    """
    In principle, create custom two-categories DataFrame with appropriate frequency and,
    if applicable, lags.

    :param <pd.Dataframe> df: standardized JPMaQS DataFrame with the following necessary
        columns: 'cid', 'xcat', 'real_date' and at least one column with values of
        interest.
    :param <List[str]> xcats: extended categories involved in the custom DataFrame. The
        last category in the list represents the dependent variable, and the (n - 1)
        preceding categories will be the explanatory variables(s).
    :param <List[str]> cids: cross-sections to be included. Default is all in the
        DataFrame.
    :param <str> val: name of column that contains the values of interest. Default is
        'value'.
    :param <str> start: earliest date in ISO 8601 format. Default is None,
        i.e. earliest date in DataFrame is used.
    :param <str> end: latest date in ISO 8601 format. Default is None,
        i.e. latest date in DataFrame is used.
    :param <dict> blacklist: cross-sections with date ranges that should be excluded from
        the DataFrame. If one cross section has several blacklist periods append numbers
        to the cross section code.
    :param <int> years: number of years over which data are aggregated. Supersedes the
        "freq" parameter and does not allow lags, Default is None, i.e. no multi-year
        aggregation.
    :param <str> freq: letter denoting frequency at which the series are to be sampled.
        This must be one of 'D', 'W', 'M', 'Q', 'A'. Default is 'M'. Will always be the
        last business day of the respective frequency.
    :param <int> lag: lag (delay of arrival) of explanatory category(s) in periods
        as set by freq. Default is 0.
    :param <int> fwin: forward moving average window of first category. Default is 1,
        i.e no average.
        Note: This parameter is used mainly for target returns as dependent variable.
    :param <List[str]> xcat_aggs: exactly two aggregation methods. Default is 'mean' for
        both. The same aggregation method, the first method in the parameter, will be
        used for all explanatory variables.

    :return <pd.DataFrame>: custom DataFrame with category columns. All rows that contain
        NaNs will be excluded.

    N.B.:
    The number of explanatory categories that can be included is not restricted and will
    be appended column-wise to the returned DataFrame. The order of the DataFrame's
    columns will reflect the order of the categories list.
    """

    freq = _map_to_business_day_frequency(freq)

    assert isinstance(xcats, list), f"<list> expected and not {type(xcats)}."
    assert all([isinstance(c, str) for c in xcats]), "List of categories expected."

    aggs_error = "List of strings, outlining the aggregation methods, expected."
    assert isinstance(xcat_aggs, list), aggs_error
    assert all([isinstance(a, str) for a in xcat_aggs]), aggs_error
    aggs_len = (
        "Only two aggregation methods required. The first will be used for all "
        "explanatory category(s)."
    )
    assert len(xcat_aggs) == 2, aggs_len

    assert not (years is not None) & (
        lag != 0
    ), "Lags cannot be applied to year groups."
    if years is not None:
        assert isinstance(start, str), "Year aggregation requires a start date."

        no_xcats = (
            "If the data is aggregated over a multi-year timeframe, only two "
            "categories are permitted."
        )
        assert len(xcats) == 2, no_xcats

    input_xcats = xcats
    input_cids = cids
    df, xcats, cids = reduce_df(df, xcats, cids, start, end, blacklist, out_all=True)

    if len(xcats) < 2:
        raise ValueError("The DataFrame must contain at least two categories. ")
    elif set(xcats) != set(input_xcats):
        missing_xcats = list(set(input_xcats) - set(xcats))
        warnings.warn(
            f"The following categories are missing from the DataFrame: {missing_xcats}"
        )

    if len(cids) < 1:
        raise ValueError(
            "The DataFrame must contain at least one valid cross section. "
        )
    elif input_cids and set(cids) != set(input_cids):
        missing_cids = list(set(input_cids) - set(cids))
        warnings.warn(
            f"The following cross sections are missing from the DataFrame: {missing_cids}"
        )

    metric = ["value", "grading", "mop_lag", "eop_lag"]
    val_error = (
        "The column of interest must be one of the defined JPMaQS metrics, "
        f"{metric}, but received {val}."
    )
    assert val in metric, val_error
    avbl_cols = list(df.columns)
    assert val in avbl_cols, (
        f"The passed column name, {val}, must be present in the "
        f"received DataFrame. DataFrame contains {avbl_cols}."
    )

    # Reduce the columns in the DataFrame to the necessary columns:
    # ['cid', 'xcat', 'real_date'] + [val] (name of column that contains the
    # values of interest: "value", "grading", "mop_lag", "eop_lag").
    col_names = ["cid", "xcat", "real_date", val]

    df_output = []
    if years is None:
        dfw = df.pivot(index=("cid", "real_date"), columns="xcat", values=val)

        dep = xcats[-1]
        # The possibility of multiple explanatory variables.
        explanatory_xcats = xcats[:-1]

        dfw = dfw.groupby(
            [
                pd.Grouper(level="cid"),
                pd.Grouper(level="real_date", freq=freq),
            ]
        )

        dfw_explanatory = _categories_df_explanatory_df(
            dfw=dfw,
            explanatory_xcats=explanatory_xcats,
            agg_method=xcat_aggs[0],
            sum_condition=(xcat_aggs[0] == "sum"),
            lag=lag,
        )

        # Handles for falsified zeros. Following the frequency conversion, if the
        # aggregation method is set to "sum", time periods that exclusively contain NaN
        # values will incorrectly be summed to the value zero which is misleading for
        # analysis.
        if not (xcat_aggs[-1] == "sum"):
            dep_col = dfw[dep].agg(xcat_aggs[1]).astype(dtype=np.float32)
        else:
            dep_col = dfw[dep].sum(min_count=1)

        if fwin > 1:
            s = 1 - fwin
            dep_col = dep_col.rolling(window=fwin).mean().shift(s)

        dfw_explanatory[dep] = dep_col
        # Order such that the return category is the right-most column - will reflect the
        # order of the categories list.
        dfc = dfw_explanatory[explanatory_xcats + [dep]]

    else:
        start_year = pd.to_datetime(start).year
        end_year = df["real_date"].max().year + 1

        grouping = int((end_year - start_year) / years)
        remainder = (end_year - start_year) % years

        year_groups = {}

        group_start_year = start_year
        for group in range(grouping):
            value = [i for i in range(group_start_year, group_start_year + years)]
            key = f"{group_start_year} - {group_start_year + (years - 1)}"
            year_groups[key] = value

            group_start_year += years

        v = [i for i in range(group_start_year, group_start_year + (remainder + 1))]
        year_groups[f"{group_start_year} - now"] = v
        list_y_groups = list(year_groups.keys())

        translate_ = lambda year: list_y_groups[int((year % start_year) / years)]
        df["real_date"] = pd.to_datetime(df["real_date"], errors="coerce")
        df["custom_date"] = df["real_date"].dt.year.apply(translate_)

        dfx_list = [df[df["xcat"] == xcats[0]], df[df["xcat"] == xcats[1]]]
        df_agg = list(map(categories_df_aggregation_helper, dfx_list, xcat_aggs))
        df_output.extend([d[col_names] for d in df_agg])

        dfc = pd.concat(df_output)
        dfc = dfc.pivot(index=("cid", "real_date"), columns="xcat", values=val)

    # Adjusted to account for multiple signals requested. If the DataFrame is
    # two-dimensional, signal & a return, NaN values will be handled inside other
    # functionality, as categories_df() is simply a support function. If the parameter
    # how is set to "any", a potential unnecessary loss of data on certain categories
    # could arise.
    return dfc.dropna(axis=0, how="all")


def years_btwn_dates(start_date: pd.Timestamp, end_date: pd.Timestamp) -> int:
    """Returns the number of years between two dates."""
    return end_date.year - start_date.year


def quarters_btwn_dates(start_date: pd.Timestamp, end_date: pd.Timestamp) -> int:
    """Returns the number of quarters between two dates."""
    return (end_date.year - start_date.year) * 4 + (
        end_date.quarter - start_date.quarter
    )


def months_btwn_dates(start_date: pd.Timestamp, end_date: pd.Timestamp) -> int:
    """Returns the number of months between two dates."""
    return (end_date.year - start_date.year) * 12 + (end_date.month - start_date.month)


def weeks_btwn_dates(start_date: pd.Timestamp, end_date: pd.Timestamp) -> int:
    """Returns the number of business weeks between two dates."""
    next_monday = start_date + pd.offsets.Week(weekday=0)
    dif = (end_date - next_monday).days // 7 + 1
    return dif


def _get_edge_dates(
    dates: Optional[Union[pd.DatetimeIndex, pd.Series, Iterable[pd.Timestamp]]] = None,
    start_date: Optional[Union[str, pd.Timestamp]] = None,
    end_date: Optional[Union[str, pd.Timestamp]] = None,
    freq: str = "M",
    direction: str = "end",
) -> pd.Series:
    assert direction in ["start", "end"], "Direction must be either 'start' or 'end'."
    datettypes = [pd.Timestamp, str, np.datetime64, datetime.date]

    freq = _map_to_business_day_frequency(freq)

    if bool(start_date) != bool(end_date):
        raise ValueError(
            "Both `start_date` and `end_date` must be passed when using "
            "dates as a start and end date."
        )

    if dates is not None:
        if not isinstance(dates, (pd.DatetimeIndex, pd.Series, Iterable)):
            raise TypeError(
                "Dates must be a pandas DatetimeIndex, Series, or a generic iterable."
            )
        if isinstance(dates, pd.DataFrame):
            dates = dates.iloc[:, 0]
        if isinstance(dates, (pd.DatetimeIndex, pd.Series)):
            dates = dates.tolist()
        dates = list(dates)

        for ix, dt in enumerate(dates):
            try:
                dates[ix] = pd.to_datetime(dt)
            except Exception as e:
                raise TypeError(
                    f"Error converting date at index {ix} to a pandas Timestamp: {e}"
                )

    if bool(start_date) and bool(dates):
        raise ValueError(
            "Only one of `dates` or `start_date` and `end_date` must be passed."
        )

    if bool(start_date):
        assert bool(end_date)
        for date, dname in zip([start_date, end_date], ["start_date", "end_date"]):
            if not isinstance(date, (str, pd.Timestamp)):
                raise TypeError(f"{dname} must be a string or a pandas Timestamp.")
            if isinstance(date, str):
                if not is_valid_iso_date(date):
                    raise ValueError(
                        "Both `start_date` and `end_date` must be valid ISO dates when passed as "
                        "strings."
                    )

        if pd.Timestamp(start_date) > pd.Timestamp(end_date):
            start_date, end_date = end_date, start_date

    dts: pd.DataFrame = pd.DataFrame(
        (
            dates
            if (dates is not None)
            else pd.bdate_range(start=start_date, end=end_date)
        ),
        columns=["real_date"],
    ).apply(pd.to_datetime, axis=1)
    min_date: pd.Timestamp = dts["real_date"].min()

    if freq == _map_to_business_day_frequency("D"):
        max_date = dts["real_date"].max()
        dtx = pd.bdate_range(start=min_date, end=max_date)
        return dtx[dtx.isin(dts["real_date"])]

    if freq == _map_to_business_day_frequency("M"):
        func = months_btwn_dates
    elif freq == _map_to_business_day_frequency("W"):
        func = weeks_btwn_dates
    elif freq == _map_to_business_day_frequency("Q"):
        func = quarters_btwn_dates
    elif freq == _map_to_business_day_frequency("A"):
        func = years_btwn_dates
    else:
        raise ValueError("Frequency parameter must be one of D, M, W, Q, or A.")

    dts["period"] = dts["real_date"].apply(func, args=(min_date,))

    dx = -1 if direction == "end" else 1
    t_indices: pd.Series = dts["period"].shift(dx) != dts["period"]
    t_dates: pd.Series = dts["real_date"].loc[t_indices].reset_index(drop=True)

    return t_dates


def get_eops(
    dates: Optional[Union[pd.DatetimeIndex, pd.Series, Iterable[pd.Timestamp]]] = None,
    start_date: Optional[Union[str, pd.Timestamp]] = None,
    end_date: Optional[Union[str, pd.Timestamp]] = None,
    freq: str = "M",
) -> pd.Series:
    """
    Returns a series of end-of-period dates for a given frequency.
    Dates can be passed as a series, index, a generic iterable or as a start and end date.

    :param <str> freq: The frequency string. Must be one of "D", "W", "M", "Q", "A".
    :param <pd.DatetimeIndex | pd.Series | Iterable[pd.Timestamp]> dates: The dates to
        be used to generate the end-of-period dates. Can be passed as a series, index, a
        generic iterable or as a start and end date.
    :param <str | pd.Timestamp> start_date: The start date. Must be passed if dates is
        not passed.
    """
    direction = "end"
    return _get_edge_dates(
        dates=dates,
        start_date=start_date,
        end_date=end_date,
        freq=freq,
        direction=direction,
    )

def merge_categories(df: pd.DataFrame, xcats: List[str], new_xcat: str, cids: List[str] = None):
    """
    Merges categories of different preferences into a single one, with the most preferred 
    being used first and others substituted in order.

    :param <pd.DataFrame> df: standardized JPMaQS DataFrame with the necessary columns
        'cid', 'xcat', 'real_date' and at least one column with values of interest.
    :param <List[str]> xcats: extended categories to be merged.
    :param <List[str]> cids: cross sections to be included. Default is all in the
        DataFrame.
    :param <str> new_xcat: name of the new category to be created. Default is None.

    :return <pd.DataFrame>: DataFrame with the merged category.
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError("The DataFrame must be a pandas DataFrame.")
    if not isinstance(xcats, list):
        raise TypeError("The categories must be a list of strings.")
    if not all(isinstance(xcat, str) for xcat in xcats):
        raise TypeError("The categories must be a list of strings.")
    if not isinstance(df, QuantamentalDataFrame):
        raise TypeError("The DataFrame must be a Quantamental DataFrame.")
    if not isinstance(new_xcat, str):
        raise TypeError("The new category must be a string.")
    if not set(xcats).issubset(df["xcat"].unique()):
        raise ValueError("The categories must be present in the DataFrame.")
    if cids is None:
        cids = list(df["cid"].unique())
    if not isinstance(cids, list):
        raise TypeError("The cross sections must be a list of strings.")
    if not all(isinstance(cid, str) for cid in cids):
        raise TypeError("The cross sections must be a list of strings.")
    if not set(cids).issubset(df["cid"].unique()):
        raise ValueError("The cross sections must be present in the DataFrame.")

    real_dates = list(df["real_date"].unique())

    def _get_values_for_xcat(real_dates, xcat_index, cid):

        values = df[(df["real_date"].isin(real_dates)) & (df["xcat"] == xcats[xcat_index]) & (df["cid"] == cid)]
        if not real_dates == list(values["real_date"].unique()):
            if xcat_index + 1 >= len(xcats):
                return values
            values = update_df(values, _get_values_for_xcat(list(set(real_dates) - set(values["real_date"].unique())), xcat_index + 1, cid))

        values.loc[:, "xcat"] = new_xcat
        return values
    
    result_df = None

    for cid in cids:
        if result_df is None:
            result_df = _get_values_for_xcat(real_dates, 0, cid)
        else:
            result_df = update_df(result_df, _get_values_for_xcat(real_dates, xcat_index=0, cid=cid))

    return result_df.sort_values(by=IDX_COLS_SORT_ORDER).reset_index(drop=True)

def get_sops(
    dates: Optional[Union[pd.DatetimeIndex, pd.Series, Iterable[pd.Timestamp]]] = None,
    start_date: Optional[Union[str, pd.Timestamp]] = None,
    end_date: Optional[Union[str, pd.Timestamp]] = None,
    freq: str = "M",
) -> pd.Series:
    """
    Returns a series of start-of-period dates for a given frequency.
    Dates can be passed as a series, index, a generic iterable or as a start and end date.

    :param <str> freq: The frequency string. Must be one of "D", "W", "M", "Q", "A".
    :param <pd.DatetimeIndex | pd.Series | Iterable[pd.Timestamp]> dates: The dates to
        be used to generate the start-of-period dates. Can be passed as a series, index, a
        generic iterable or as a start and end date.
    :param <str | pd.Timestamp> start_date: The start date. Must be passed if dates is
        not passed.
    """
    direction = "start"
    return _get_edge_dates(
        dates=dates,
        start_date=start_date,
        end_date=end_date,
        freq=freq,
        direction=direction,
    )


if __name__ == "__main__":
    from macrosynergy.management.simulate import make_qdf

    cids = ["AUD", "CAD", "GBP", "NZD"]
    xcats = ["XR1", "XR2", "CRY1", "CRY2"]
    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2000-01-01", "2020-12-31", 0.1, 1]
    df_cids.loc["CAD"] = ["2001-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP"] = ["2002-01-01", "2020-11-30", 0, 2]
    df_cids.loc["NZD"] = ["2002-01-01", "2020-09-30", -0.1, 2]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR1"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["XR2"] = ["2000-01-01", "2020-10-30", 1, 2, 0.95, 1]
    df_xcats.loc["CRY1"] = ["2001-01-01", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["CRY2"] = ["2001-01-01", "2020-10-30", 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    dfw = categories_df(
        df=dfd,
        xcats=xcats[:2] + ["test"],
        cids=cids,
        freq="M",
        # lag=1,
        xcat_aggs=["last", "sum"],
        # years=5,
        # start="2000-01-01",
    )



from .basket import Basket
from .category_relations import CategoryRelations
from .converge_row import ConvergeRow
from .return_beta import return_beta, beta_display
from .historic_vol import historic_vol
from .linear_composite import linear_composite
from .make_blacklist import make_blacklist
from .make_relative_value import make_relative_value
from .make_relative_category import make_relative_category
from .make_zn_scores import make_zn_scores
from .panel_calculator import panel_calculator
from .view_correlations import correl_matrix
from .view_grades import heatmap_grades
from .view_ranges import view_ranges
from .view_timelines import view_timelines
from .view_metrics import view_metrics

__all__ = [
    "Basket",
    "CategoryRelations",
    "ConvergeRow",
    "return_beta",
    "beta_display",
    "historic_vol",
    "linear_composite",
    "make_blacklist",
    "make_relative_value",
    "make_relative_category",
    "make_zn_scores",
    "panel_calculator",
    "correl_matrix",
    "heatmap_grades",
    "view_ranges",
    "view_timelines",
    "view_metrics",
]



"""
Module for checking the availability of data availabity from a 
Quantamental DataFrame. Includes functions for checking start years
and end dates of a DataFrame, as well as visualizing the results.
"""

import numpy as np
import pandas as pd
from typing import List, Tuple

from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import reduce_df
from macrosynergy.management.types import QuantamentalDataFrame
import macrosynergy.visuals as msv


def check_availability(
    df: pd.DataFrame,
    xcats: List[str] = None,
    cids: List[str] = None,
    start: str = None,
    start_size: Tuple[float] = None,
    end_size: Tuple[float] = None,
    start_years: bool = True,
    missing_recent: bool = True,
    use_last_businessday: bool = True,
):
    """
    Wrapper for visualizing start and end dates of a filtered DataFrame.

    :param <pd.DataFrame> df: standardized DataFrame with the following necessary
        columns: 'cid', 'xcat', 'real_date'.
    :param <List[str]> xcats: extended categories to be checked on.
        Default is all in the DataFrame.
    :param <List[str]> cids: cross sections to be checked on.
        Default is all in the DataFrame.
    :param <str> start: string representing earliest considered date. Default is None.
    :param <Tuple[float]> start_size: tuple of floats with width / length of
        the start years heatmap. Default is None (format adjusted to data).
    :param <Tuple[float]> end_size: tuple of floats with width/length of
        the end dates heatmap. Default is None (format adjusted to data).
    :param <bool> start_years: boolean indicating whether or not to display a chart
        of starting years for each cross-section and indicator.
        Default is True (display start years).
    :param <bool> missing_recent: boolean indicating whether or not to display a chart
        of missing date numbers for each cross-section and indicator.
        Default is True (display missing days).
    :param <bool> use_last_businessday: boolean indicating whether or not to use the
        last business day before today as the end date. Default is True.
    """
    if not isinstance(start_years, bool):
        raise TypeError(f"<bool> object expected and not {type(start_years)}.")
    if not isinstance(missing_recent, bool):
        raise TypeError(f"<bool> object expected and not {type(missing_recent)}.")

    dfx = reduce_df(df, xcats=xcats, cids=cids, start=start)
    if start_years:
        dfs = check_startyears(dfx)
        visual_paneldates(
            dfs, size=start_size, use_last_businessday=use_last_businessday
        )
    if missing_recent:
        dfe = check_enddates(dfx)
        visual_paneldates(dfe, size=end_size, use_last_businessday=use_last_businessday)


def missing_in_df(
    df: QuantamentalDataFrame,
    xcats: List[str] = None,
    cids: List[str] = None,
):
    """
    Print missing cross-sections and categories

    :param <QuantamentalDataFrame> df: standardized DataFrame with the following necessary
        columns: 'cid', 'xcat', 'real_date'.
    :param <List[str]> xcats: extended categories to be checked on. Default is all
        in the DataFrame.
    :param <List[str]> cids: cross sections to be checked on. Default is all in
        the DataFrame.

    """
    if not isinstance(df, QuantamentalDataFrame):
        raise TypeError("`df` must be a QuantamentalDataFrame/pd.DataFrame")

    if df.empty:
        raise ValueError("`df` is empty.")

    for lst, name in zip([xcats, cids], ["xcats", "cids"]):
        if (lst is not None) and not (
            isinstance(lst, list) and all(isinstance(x, str) for x in lst)
        ):
            raise TypeError(f"`{name}` should be a `List[str]` and not {type(lst)}.")

    missing_across_df = list(set(xcats) - set(df["xcat"]))
    if len(missing_across_df) > 0:
        print("Missing XCATs across DataFrame: ", missing_across_df)
    else:
        print("No missing XCATs across DataFrame.")

    cids = df["cid"].unique() if cids is None else cids
    xcats_used = sorted(list(set(xcats).intersection(set(df["xcat"]))))
    if len(xcats_used) == 0:
        print("No XCATs found in the DataFrame.")
        return

    max_xcat_len = max(map(len, xcats_used))
    for xcat in xcats_used:
        cids_xcat = df.loc[df["xcat"] == xcat, "cid"].unique()
        missing_cids = sorted(set(cids) - set(cids_xcat))
        msg = f"Missing cids for {xcat}: " + " " * (max_xcat_len - len(xcat))
        print(msg, missing_cids)


def check_startyears(df: pd.DataFrame):
    """
    DataFrame with starting years across all extended categories and cross-sections

    :param <pd.DataFrame> df: standardized DataFrame with the following necessary
        columns: 'cid', 'xcat', 'real_date'.

    """
    df: pd.DataFrame = df.copy()
    df = df.dropna(how="any")
    df_starts = df[["cid", "xcat", "real_date"]].groupby(["cid", "xcat"]).min()
    df_starts["real_date"] = pd.DatetimeIndex(df_starts.loc[:, "real_date"]).year

    return df_starts.unstack().loc[:, "real_date"].astype(int, errors="ignore")


def check_enddates(df: pd.DataFrame) -> pd.DataFrame:
    """
    DataFrame with end dates across all extended categories and cross sections.

    :param <pd.DataFrame> df: standardized DataFrame with the following necessary
        columns: 'cid', 'xcat', 'real_date'.
    """
    df: pd.DataFrame = df.copy()
    df = df.dropna(how="any")
    df_ends = df[["cid", "xcat", "real_date"]].groupby(["cid", "xcat"]).max()
    df_ends["real_date"] = df_ends["real_date"].dt.strftime("%Y-%m-%d")

    return df_ends.unstack().loc[:, "real_date"]


def business_day_dif(df: pd.DataFrame, maxdate: pd.Timestamp):
    """
    Number of business days between two respective business dates.

    :param <pd.DataFrame> df: DataFrame cross-sections rows and category columns. Each
        cell in the DataFrame will correspond to the start date of the respective series.
    :param <pd.Timestamp> maxdate: maximum release date found in the received DataFrame.
        In principle, all series should have values up until the respective business
        date. The difference will represent possible missing values.

    :return <pd.DataFrame>: DataFrame consisting of business day differences for all
        series.

    """
    year_df = (maxdate.year - df.apply(lambda x: x.dt.isocalendar().year)) * 52
    week_df = maxdate.week - df.apply(lambda x: x.dt.isocalendar().week)
    # Account for difference over a year.
    week_df += year_df
    # Account for weekends.
    week_df *= 2
    df = (maxdate - df).apply(lambda x: x.dt.days)
    # set to zero if the difference is negative.
    df = df - week_df
    return df.where(df >= 0, 0)


def visual_paneldates(
    df: pd.DataFrame, size: Tuple[float] = None, use_last_businessday: bool = True
):
    """
    Visualize panel dates with color codes.

    :param <pd.DataFrame> df: DataFrame cross sections rows and category columns.
    :param <Tuple[float]> size: tuple of floats with width/length of displayed heatmap.
    :param <bool> use_last_businessday: boolean indicating whether or not to use the
        last business day before today as the end date. Default is True.

    """
    msv.view_panel_dates(df=df, size=size, use_last_businessday=use_last_businessday)


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP"]
    xcats = ["XR", "CRY"]

    cols_1 = ["earliest", "latest", "mean_add", "sd_mult"]
    df_cids = pd.DataFrame(index=cids, columns=cols_1)
    df_cids.loc["AUD",] = ["2010-01-01", "2020-12-31", 0.5, 2]
    df_cids.loc["CAD",] = ["2010-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP",] = ["2012-01-01", "2020-11-30", -0.2, 0.5]

    cols_2 = cols_1 + ["ar_coef", "back_coef"]
    df_xcats = pd.DataFrame(index=xcats, columns=cols_2)
    df_xcats.loc["XR",] = ["2010-01-01", "2020-12-31", 0, 1, 0, 0.3]
    df_xcats.loc["CRY",] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    filt_na = (dfd["cid"] == "CAD") & (dfd["real_date"] < "2011-01-01")
    dfd.loc[filt_na, "value"] = np.nan

    xxcats = xcats + ["TREND"]
    xxcids = cids + ["USD"]

    check_availability(
        df=dfd, xcats=xcats, cids=cids, start_size=(10, 5), end_size=(10, 8)
    )



"""Macrosynergy Quant Research Package"""

try:
    from .version import git_revision as __git_revision__
    from .version import version as __version__
except ModuleNotFoundError:
    from setup import get_version_info

    FULLVERSION, GIT_REVISION = get_version_info()
    __version__ = FULLVERSION
    __git_revision__ = GIT_REVISION

from . import visuals, download, panel, pnl, management, signal, learning
from .management.utils import check_package_version

__all__ = [
    "visuals",
    "download",
    "panel",
    "pnl",
    "management",
    "signal",
    "learning",
    "check_package_version"
]

__name__ = ["__version__"]


# allows the package version information to be accessed from the package



"""
Implementation of panel calculation functions for quantamental data.
The functionality allows applying mathematical operations on time-series data.

"""

import numpy as np
import pandas as pd
from typing import List, Tuple
from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import reduce_df
from macrosynergy.management.utils import drop_nan_series
import re
import random


def time_series_check(formula: str, index: int):
    """
    Determine if the panel has any time-series methods applied. If a time-series
    conversion is applied, the function will return the terminal index of the respective
    category. Further, a boolean parameter is also returned to confirm the presence of a
    time-series operation.

    :param <str> formula:
    :param <int> index: starting index to iterate over.

    :return <Tuple[int, bool]>:
    """

    check = lambda a, b, c: (
        (a.isupper() or a.isnumeric()) and b == "." and c.islower()
    )

    f = formula
    length = len(f)
    clause = False
    for i in range(index, (length - 2)):
        if check(f[i], f[i + 1], f[i + 2]):
            clause = True
            break

    return i, clause


def xcat_isolator(expression: str, start_index: str, index: int):
    """
    Split the category from the time-series operation. The function will return the
    respective category.

    :param <str> expression:
    :param <str> start_index: starting index to search over.
    :param <int> index: defines the end of the search space over the expression.

    :return <str>: xcat.
    """

    op_copy = expression[start_index : index + 1]

    start = next(i for i, elem in enumerate(op_copy) if elem.isupper())

    xcat = op_copy[start : index + 1]

    return xcat, start_index + start + len(xcat)


def _get_xcats_used(ops: dict) -> Tuple[List[str], List[str]]:
    """
    Collect all categories used in the panel calculation.

    :param <dict> ops: dictionary of panel calculation formulas.

    :return <Tuple[List[str], List[str]]>: all_xcats_used, singles_used.
    """
    xcats_used: List[str] = []
    singles_used: List[str] = []
    for op in ops.values():
        index, clause = time_series_check(formula=op, index=0)
        start_index = 0
        if clause:
            while clause:
                xcat, end_ = xcat_isolator(op, start_index, index)
                xcats_used.append(xcat)
                index, clause = time_series_check(op, index=end_)
                start_index = end_
        else:
            op_list = op.split(" ")
            xcats_used += [x for x in op_list if re.match("^[A-Z]", x)]
            singles_used += [s for s in op_list if re.match("^i", s)]

    single_xcats = [x[5:] for x in singles_used]
    single_cids = [x[1:4] for x in single_xcats]
    all_xcats_used = xcats_used + single_xcats
    return all_xcats_used, singles_used, single_cids


def panel_calculator(
    df: pd.DataFrame,
    calcs: List[str] = None,
    cids: List[str] = None,
    start: str = None,
    end: str = None,
    blacklist: dict = None,
    external_func: dict = None
) -> pd.DataFrame:
    """
    Calculates new data panels through operations on existing panels.

    :param <pd.Dataframe> df: standardized dataframe with following necessary columns:
        'cid', 'xcat', 'real_date' and 'value'.
    :param <List[str]> calcs:  list of formulas denoting operations on panels of
        categories. Words in capital letters denote category panels.
        Otherwise the formulas can include numpy functions and standard binary operators.
        See notes below.
    :param <List[str]> cids: cross sections over which the panels are defined.
    :param <str> start: earliest date in ISO format. Default is None and earliest date in
        df is used.
    :param <str> end: latest date in ISO format. Default is None and latest date in df is
        used.
    :param <dict> blacklist: cross sections with date ranges that should be excluded from
        the dataframe. If one cross section has several blacklist periods append numbers
        to the cross-section code.
    :param <dict> external_func: dictionary of external functions to be used in the panel
        calculation. The key is the name of the function and the value is the function.

    :return <pd.Dataframe>: standardized dataframe with all new categories in standard
        format, i.e the columns 'cid', 'xcat', 'real_date' and 'value'.

    Notes:
    Panel calculation strings can use numpy functions and unary/binary operators on
    category panels. The category is indicated by capital letters, underscores
    and numbers.
    Panel category names that are not at the beginning or end of the string must always
    have a space before and after the name.
    Calculated category and panel operations must be separated by '='. Examples:
        "NEWCAT = ( OLDCAT1 + 0.5) * OLDCAT2"
        "NEWCAT = np.log( OLDCAT1 ) - np.abs( OLDCAT2 ) ** 1/2"
    Panel calculation can also involve individual indicator series (to be applied
    to all series in the panel by using th 'i' as prefix), such as:
        "NEWCAT = OLDCAT1 - np.sqrt( iUSD_OLDCAT2 )"
    If more than one new category is calculated, the resulting panels can be used
    sequentially in the calculations, such as:
        ["NEWCAT1 = 1 + OLDCAT1 / 100", "NEWCAT2 = OLDCAT2 * NEWCAT1"]
    """

    # A. Asserts

    cols = ["cid", "xcat", "real_date", "value"]

    col_error = f"The DataFrame must contain the necessary columns: {cols}."
    assert set(cols).issubset(set(df.columns)), col_error
    # Removes any columns beyond the required.
    df = df.loc[:, cols]

    df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")

    assert isinstance(calcs, list), "List of functions expected."

    error_formula = "Each formula in the panel calculation list must be a string."
    assert all([isinstance(elem, str) for elem in calcs]), error_formula
    assert isinstance(cids, list), "List of cross-sections expected."

    _check_calcs(calcs)

    if external_func:
        globals().update(external_func)

    # B. Collect new category names and their formulas.

    ops = {}
    for calc in calcs:
        calc_parts = calc.split("=", maxsplit=1)
        ops[calc_parts[0].strip()] = calc_parts[1].strip()

    # C. Check if all required categories are in the dataframe.

    old_xcats_used, singles_used, single_cids = _get_xcats_used(ops)

    old_xcats_used = list(set(old_xcats_used))
    missing = sorted(set(old_xcats_used) - set(df["xcat"].unique()))

    new_xcats = list(ops.keys())
    if len(missing) > 0 and not set(missing).issubset(set(new_xcats)):
        raise ValueError(f"Missing categories: {missing}.")

    # If any of the elements of single_cids are not in cids, add them to cids.
    cids_used = list(set(single_cids + cids))

    # D. Reduce dataframe with intersection requirement.

    dfx = reduce_df(
        df,
        xcats=old_xcats_used,
        cids=cids_used,
        start=start,
        end=end,
        blacklist=blacklist,
        intersect=False,
    )

    # E. Create all required wide dataframes with category names.

    for xcat in old_xcats_used:
        dfxx = dfx[dfx["xcat"] == xcat]
        dfw = dfxx.pivot(index="real_date", columns="cid", values="value")
        exec(f"{xcat} = dfw")

    for single in singles_used:
        ticker = single[1:]
        dfxx = df[(df["cid"] + "_" + df["xcat"]) == ticker]
        if dfxx.empty:
            raise ValueError(f"Ticker, {ticker}, missing from the dataframe.")
        else:
            dfx1 = dfxx.set_index("real_date")["value"].to_frame()
            dfx1 = dfx1.truncate(before=start, after=end)

            dfw = pd.concat([dfx1] * len(cids), axis=1, ignore_index=True)
            dfw.columns = cids
            exec(f"{single} = dfw")

    # F. Calculate the panels and collect.
    df_out: pd.DataFrame
    for new_xcat, formula in ops.items():
        dfw_add = eval(formula)
        df_add = pd.melt(dfw_add.reset_index(), id_vars=["real_date"]).rename(
            {"variable": "cid"}, axis=1
        )
        df_add["xcat"] = new_xcat
        if new_xcat == list(ops.keys())[0]:
            df_out = df_add[cols]
        else:
            df_out = pd.concat([df_out, df_add[cols]], axis=0, ignore_index=True)
        exec(f"{new_xcat} = dfw_add")

    if df_out.isna().any().any():
        df_out = drop_nan_series(df=df_out, raise_warning=True)

    return df_out


def _check_calcs(formulas: List[str]):
    """
    Check formulas for invalid characters in xcats.

    :param <List[str]> calcs: list of formulas.

    :return <List[str]>: list of formulas.
    """
    pattern = r"[-+*()/](?=i?[A-Z])|(?<=[A-Z])[-+*()/]"

    for formula in formulas:
        for term in formula.split():
            # Search for any occurrences of the pattern in the input string
            if re.search(pattern, term):
                raise ValueError(
                    f"Invalid character found next to a capital letter or 'i' in string: {term}. "
                    + "Arithmetic operators and parentheses must be separated by spaces."
                )


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "USD", "NZD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )

    df_cids.loc["AUD"] = ["2010-01-01", "2020-12-31", 0.5, 2]
    df_cids.loc["CAD"] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP"] = ["2012-01-01", "2020-11-30", -0.2, 0.5]
    df_cids.loc["USD"] = ["2010-01-01", "2020-12-30", -0.2, 0.5]
    df_cids.loc["NZD"] = ["2002-01-01", "2020-09-30", -0.1, 2]
    df_cids.loc["EUR"] = ["2002-01-01", "2020-09-30", -0.2, 2]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR"] = ["2012-01-01", "2020-12-31", 0, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2010-01-01", "2020-10-30", 1, 2, 0.9, 0.5]
    df_xcats.loc["GROWTH"] = ["2012-01-01", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2012-01-01", "2020-09-30", 1, 2, 0.8, 0.5]

    random.seed(2)
    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    # Example blacklist.
    black = {"AUD": ["2000-01-01", "2003-12-31"]}

    start = "2010-01-01"
    end = "2020-12-31"

    # Example filter for dataframe.
    filt1 = (dfd["xcat"] == "XR") | (dfd["xcat"] == "CRY")
    dfdx = dfd[filt1]

    # First testcase.

    f1 = "NEW_VAR1 = GROWTH - iEUR_INFL"
    formulas = [f1]
    cidx = ["AUD", "CAD"]
    df_calc = panel_calculator(
        df=dfd, calcs=formulas, cids=cidx, start=start, end=end, blacklist=black
    )
    # Second testcase: EUR is not passed in as one of the cross-sections in "cids"
    # parameter but is defined in the dataframe. Therefore, code will not break.
    cids = ["AUD", "CAD", "GBP", "USD", "NZD"]
    formula = "NEW1 = XR - iUSD_XR"
    formula_2 = "NEW2 = GROWTH - iEUR_INFL"
    formulas = [formula, formula_2]
    df_calc = panel_calculator(
        df=dfd, calcs=formulas, cids=cids, start=start, end=end, blacklist=black
    )



"""
Module housing decorators that are used to validate the arguments and return values of
functions.
"""

import inspect
import warnings
from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
)
from inspect import signature
from macrosynergy.management.types import NoneType
import pandas as pd
import numpy as np
from packaging import version

try:
    from macrosynergy import __version__ as _version
except ImportError:
    try:
        from setup import VERSION as _version
    except ImportError:
        _version = "0.0.0"


def deprecate(
    new_func: Callable,
    deprecate_version: str,
    remove_after: str = None,
    message: str = None,
):
    """
    Decorator for deprecating a function.

    Parameters
    :param <callable> new_func: The function that replaces the old one.
    :param <str> deprecate_version: The version in which the old function is deprecated.
    :param <str> remove_after: The version in which the old function is removed.
    :param <str> message: The message to display when the old function is called.
        This message must contain the following format strings:
        "{old_method}", "{deprecate_version}", and "{new_method}".
        If None, the default message is used.
    :return <callable>: The decorated function.
    """

    def decorator(
        old_func,
        new_func=new_func,
        deprecate_version=deprecate_version,
        remove_after=remove_after,
        message=message,
    ):
        # if the message is none, use the default message
        if message is None:
            message = (
                "{old_method} was deprecated in version {deprecate_version} and will be "
                "removed in version. Use {new_method} instead."
            )
        # else if the message does not have "{old_method}" in it, fail
        else:
            if any(
                [
                    fs not in message
                    for fs in ["{old_method}", "{deprecate_version}", "{new_method}"]
                ]
            ):
                raise ValueError(
                    "The message must contain the following format strings: "
                    "'{old_method}', '{deprecate_version}', and '{new_method}'."
                )

        # if remove_after is not None, check the version
        if remove_after is not None:
            try:
                version.parse(remove_after)
            except:
                raise ValueError(
                    f"The version in which the function is deprecated ({remove_after}) "
                    f"must be a valid version string."
                )

            if version.parse(deprecate_version) < version.parse(remove_after):
                raise ValueError(
                    f"The version in which the old function will be removed "
                    f"({remove_after}) "
                    f"must be greater than the version in which it is deprecated "
                    f"({deprecate_version})."
                )
        else:
            remove_after = _version

        @wraps(old_func)
        # This will ensure the old function retains its name and other properties.
        def wrapper(*args, **kwargs):
            if version.parse(deprecate_version) < version.parse(_version):
                warnings.warn(
                    message.format(
                        old_method=old_func.__name__,
                        new_method=new_func.__name__,
                        deprecate_version=deprecate_version,
                    ),
                    FutureWarning,
                )

            return old_func(*args, **kwargs)

        # Update the signature and docstring of the old function to match the new one.
        wrapper.__signature__ = signature(new_func)
        wrapper.__doc__ = new_func.__doc__

        return wrapper

    return decorator


def is_matching_subscripted_type(value: Any, type_hint: Type[Any]) -> bool:
    """
    Implementation of `insinstance()` for type-hints imported from the `typing` module,
    and for subscripted types (e.g. `List[int]`, `Tuple[str, int]`, etc.).
    Parameters
    :param <Any> value: The value to check.
    :param <Type[Any]> type_hint: The type hint to check against.
    :return <bool>: True if the value is of the type hint, False otherwise.
    """
    origin = get_origin(type_hint)
    args = get_args(type_hint)

    # handling lists
    if origin in [list, List]:
        if not isinstance(value, list):
            return False
        return all(isinstance(item, args[0]) for item in value)

    # tuples
    if origin in [tuple, Tuple]:
        if not isinstance(value, tuple) or len(value) != len(args):
            return False
        # don't switch order of get_origin and is_matching_subscripted_type, is
        # short-circuiting
        return all(
            [
                (get_origin(expected) and is_matching_subscripted_type(item, expected))
                or isinstance(item, expected)
                for item, expected in zip(value, args)
            ]
        )

    # dicts
    if origin in [dict, Dict]:
        if not isinstance(value, dict):
            return False
        key_type, value_type = args
        return all(
            [
                (get_origin(key_type) and is_matching_subscripted_type(k, key_type))
                or isinstance(k, key_type)
                or (isinstance(k, key_type) and isinstance(v, value_type))
                for k, v in value.items()
            ]
        )

    # unions and optionals
    if origin is Union:
        for possible_type in args:
            if get_origin(possible_type):  # is subscripted
                if is_matching_subscripted_type(value, possible_type):
                    return True
            elif isinstance(value, possible_type):
                return True
        return False

    return False


def get_expected_type(arg_type_hint: Type[Any]) -> List[str]:
    """
    Based on the type hint, return a list of strings that represent
    the type hint - including any nested type hints.
    Parameters
    :param <Type[Any]> arg_type_hint: The type hint to get the expected types for.
    :return <List[str]>: A list of strings that represent the type hint.
    """
    origin = get_origin(arg_type_hint)
    args = get_args(arg_type_hint)

    # handling lists
    if origin in [list, List]:
        return [f"List[{get_expected_type(args[0])[0]}]"]

    # tuples
    if origin in [tuple, Tuple]:
        return [f"Tuple[{', '.join(get_expected_type(arg) for arg in args)}]"]

    # dicts
    if origin in [dict, Dict]:
        return [f"Dict[{', '.join(get_expected_type(arg) for arg in args)}]"]

    # unions and optionals
    if origin in [Union, Optional]:
        # get a flat list of all the expected types
        expected_types: List[str] = []
        for possible_type in args:
            if get_origin(possible_type):
                expected_types.extend(get_expected_type(possible_type))
            else:
                expected_types.append(str(possible_type))
        return expected_types

    return [str(arg_type_hint)]


def argvalidation(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Decorator for validating the arguments and return value of a function.
    Parameters
    :param <Callable[..., Any]> func: The function to validate.
    :return <Callable[..., Any]>: The decorated function.
    """

    def format_expected_type(expected_types: List[Any]) -> str:
        # format the expected types to read nicely, and to remove 'typing.' from the string
        if isinstance(expected_types, tuple):
            expected_types = list(expected_types)
        for i, et in enumerate(expected_types):
            if str(et).startswith("typing."):
                expected_types[i] = str(et).replace("typing.", "")
            if et is NoneType:
                expected_types[i] = "None"

        ret_string = (
            f"{', '.join([f'`{t}`' for t in expected_types[:-1]])}, "
            f"or `{expected_types[-1]}`"
        )
        if len(expected_types) == 1:
            return f"`{expected_types[0]}`"
        elif len(expected_types) == 2:
            return f"`{expected_types[0]}` or `{expected_types[1]}`"
        else:
            return ret_string

    @wraps(func)
    def validation_wrapper(*args: Any, **kwargs: Any) -> Any:
        func_sig: inspect.Signature = inspect.signature(func)
        func_params: Dict[str, inspect.Parameter] = func_sig.parameters
        func_annotations: Dict[str, Any] = func_sig.return_annotation
        func_args: Dict[str, Any] = inspect.getcallargs(func, *args, **kwargs)

        # validate the arguments
        for arg_name, arg_value in func_args.items():
            if arg_name in func_params:
                arg_type: Type[Any] = func_params[arg_name].annotation
                if arg_type is not inspect._empty:
                    origin = get_origin(arg_type)
                    if origin:  # Handling subscripted types
                        # replace 'float' with 'typng.Union[float, int]' to make life easier
                        if not is_matching_subscripted_type(arg_value, arg_type):
                            exp_types: str = format_expected_type(get_args(arg_type))
                            raise TypeError(
                                f"Argument `{arg_name}` must be of type {exp_types}, "
                                f"not `{type(arg_value).__name__}` (with value "
                                f"`{arg_value}`)."
                            )
                    else:  # For simple, non-generic types
                        if not isinstance(arg_value, arg_type):
                            raise TypeError(
                                f"Argument `{arg_name}` must be of type `{arg_type}`, "
                                f"not `{type(arg_value).__name__}` (with value "
                                f"`{arg_value}`)."
                            )

        # validate the return value
        return_value: Any = func(*args, **kwargs)
        if func_annotations is not inspect._empty:
            origin = get_origin(func_annotations)
            if (
                origin
                and (not is_matching_subscripted_type(return_value, func_annotations))
            ) or (not origin and not isinstance(return_value, func_annotations)):
                exp_types: str = format_expected_type(get_args(func_annotations))
                raise warnings.warn(
                    f"Return value of `{func.__name__}` is not of type "
                    f"`{func_annotations}`, but of type `{type(return_value)}`."
                )

        return return_value

    return validation_wrapper


def argcopy(func: Callable) -> Callable:
    """
    Decorator for applying a "pass-by-value" method to the arguments of a function.
    Parameters
    :param <Callable> func: The function to copy arguments for.
    :return <Callable>: The decorated function.
    """

    @wraps(func)
    def copy_wrapper(*args, **kwargs):
        copy_types = (
            list,
            dict,
            np.ndarray,
            pd.Series,
            set,
        )
        new_args: List[Tuple[Any, ...]] = []
        for arg in args:
            if isinstance(arg, copy_types) or issubclass(type(arg), copy_types):
                new_args.append(arg.copy())
            else:
                new_args.append(arg)
        new_kwargs: Dict[str, Any] = {}
        for key, value in kwargs.items():
            if isinstance(value, copy_types) or issubclass(type(value), copy_types):
                new_kwargs[key] = value.copy()
            else:
                new_kwargs[key] = value

        return func(*new_args, **new_kwargs)

    return copy_wrapper
